"""
Analysis of currents in response to moving edges.

Info:
    Relies on dataclass defined in `flyvis.analysis.stimulus_responses_currents`.
"""

from __future__ import annotations

from typing import Any, Callable, Iterable, List, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from datamate import Namespace
from matplotlib.axes import Axes
from matplotlib.colors import hex2color
from matplotlib.lines import Line2D
from toolz import valfilter, valmap

from flyvis.connectome import ReceptiveFields
from flyvis.datasets.moving_bar import MovingEdge
from flyvis.utils.color_utils import (
    ND,
    PD,
    adapt_color_alpha,
    cmap_iter,
    truncate_colormap,
)
from flyvis.utils.df_utils import where_dataframe as get_stimulus_index
from flyvis.utils.nodes_edges_utils import CellTypeArray

from .stimulus_responses_currents import ExperimentData
from .visualization import plots, plt_utils
from .visualization.figsize_utils import (
    cm_to_inch,
    figsize_from_n_items,
)

__all__ = ["MovingEdgeCurrentView"]


class MovingEdgeCurrentView:
    """Represents a view of moving edge currents for analysis and visualization.

    This class provides methods for analyzing and visualizing currents and responses
    related to moving edge stimuli in neural simulations.

    Args:
        ensemble: The ensemble of models.
        target_type: The type of target cell.
        exp_data: Experimental data.
        arg_df: DataFrame containing stimulus arguments.
        currents: Currents for each source type.
        rfs: Receptive fields for the target cells.
        time: Time array for the simulation.
        responses: Responses of the target cells.

    Attributes:
        target_type: The type of target cell.
        ensemble: The ensemble of models.
        config: Configuration settings.
        arg_df: DataFrame containing stimulus arguments.
        rfs: Receptive fields for the target cells.
        exp_data: Experimental data.
        source_types: Types of source cells.
        time: Time array for the simulation.
        currents: Currents for each source type.
        responses: Responses of the target cells.

    Note:
        This class is intended to be updated to use xarray datasets in the future.
    """

    def __init__(
        self,
        ensemble,
        target_type: str,
        exp_data: List[ExperimentData],
        arg_df: pd.DataFrame | None = None,
        currents: Namespace | None = None,
        rfs: ReceptiveFields | None = None,
        time: np.ndarray | None = None,
        responses: np.ndarray | None = None,
    ):
        self.target_type = target_type
        self.ensemble = ensemble
        self.config = exp_data[0].config
        if arg_df is None:
            self.arg_df = MovingEdge(**exp_data[0].config).arg_df
        else:
            self.arg_df = arg_df
        self.rfs = rfs or reset_index(
            ReceptiveFields(target_type, ensemble[0].connectome.edges.to_df())
        )
        self.exp_data = exp_data
        self.source_types = self.rfs.source_types
        self.time = time
        self.init_currents(currents)
        self.init_time(time)
        self.init_responses(responses)

    def init_currents(self, currents: Namespace | None) -> None:
        """Initialize the currents for each source type.

        Args:
            currents: Currents for each source type.
        """
        if currents is not None:
            self.currents = currents
            return
        self.currents = Namespace()
        for source_type in self.rfs.source_types:
            # (on/off, n_models, n_angles, n_timesteps, n_input_cells)
            self.currents[source_type] = np.array(
                [
                    np.array(exp.target_data[self.target_type].source_data[source_type])
                    for exp in self.exp_data
                ],
            )

    def init_responses(self, responses: np.ndarray | None) -> None:
        """Initialize the responses of the target cells.

        Args:
            responses: Responses of the target cells.
        """
        if responses is not None:
            self.responses = responses
            return
        # (on/off, n_models, n_angles, n_timesteps)
        self.responses = np.array(
            [
                np.array(exp.target_data[self.target_type].activity_central)
                for exp in self.exp_data
            ],
        )

    def init_time(self, time: np.ndarray | None) -> None:
        """Initialize the time array for the simulation.

        Args:
            time: Time array for the simulation.
        """
        if time is not None:
            self.time = time
            return
        self.time = self.time or (
            np.arange(0, next(iter(self.currents.values())).shape[-2]) * self.config.dt
            - self.config.t_pre
        )

    @property
    def on(self) -> "MovingEdgeCurrentView":
        """Return a view of the ON responses."""
        on_index = get_stimulus_index(self.arg_df, intensity=0)
        arg_df = self.arg_df.iloc[on_index]
        return self.view(
            Namespace({
                cell_type: np.take(c, indices=on_index, axis=1)
                for cell_type, c in self.currents.items()
            }),
            responses=np.take(self.responses, indices=on_index, axis=1),
            arg_df=arg_df,
        )

    @property
    def off(self) -> "MovingEdgeCurrentView":
        """Return a view of the OFF responses."""
        off_index = get_stimulus_index(self.arg_df, intensity=0)
        arg_df = self.arg_df.iloc[off_index]
        return self.view(
            Namespace({
                cell_type: np.take(c, indices=off_index, axis=1)
                for cell_type, c in self.currents.items()
            }),
            responses=np.take(self.responses, indices=off_index, axis=1),
            arg_df=arg_df,
        )

    def divide_by_given_norm(self, norm: CellTypeArray) -> "MovingEdgeCurrentView":
        """Divide currents and responses by a given norm.

        Args:
            norm: The norm to divide by.

        Returns:
            A new view with normalized currents and responses.

        Raises:
            ValueError: If norm is not a CellTypeArray.
        """
        if not isinstance(norm, CellTypeArray):
            raise ValueError

        response_dims = np.arange(len(self.responses.shape))
        response_norm = np.expand_dims(
            norm[self.target_type].squeeze(), list(set(response_dims) - set([0]))
        )

        # divide the responses by the norm
        new_responses = self.responses[:] / response_norm

        # note: we also divide by the norm of the target cell type

        currents_dims = np.arange(len(next(iter(self.currents.values())).shape))

        currents_norm = np.expand_dims(
            norm[self.target_type].squeeze(), list(set(currents_dims) - set([0]))
        )

        # divide the currents by the norm
        new_currents = Namespace({
            cell_type: c / currents_norm for cell_type, c in self.currents.items()
        })
        return self.view(currents=new_currents, responses=new_responses)

    def at_contrast(self, contrast: float) -> "MovingEdgeCurrentView":
        """Create a new view filtered by contrast.

        Args:
            contrast: The contrast value to filter by.

        Returns:
            A new view with data filtered by the specified contrast.
        """
        contrast_index = get_stimulus_index(self.arg_df, intensity=contrast)
        arg_df = self.arg_df.iloc[contrast_index]
        return self.view(
            Namespace({
                cell_type: np.take(c, indices=contrast_index, axis=1)
                for cell_type, c in self.currents.items()
            }),
            responses=np.take(self.responses, indices=contrast_index, axis=1),
            arg_df=arg_df,
        )

    def at_angle(self, angle: float) -> "MovingEdgeCurrentView":
        """Create a new view filtered by angle.

        Args:
            angle: The angle value to filter by.

        Returns:
            A new view with data filtered by the specified angle.
        """
        angle_index = get_stimulus_index(self.arg_df, angle=angle)
        arg_df = self.arg_df.iloc[angle_index]
        return self.view(
            Namespace({
                cell_type: np.take(c, indices=angle_index, axis=1)
                for cell_type, c in self.currents.items()
            }),
            responses=np.take(self.responses, indices=angle_index, axis=1),
            arg_df=arg_df,
        )

    def at_position(
        self, u: float | None = None, v: float | None = None, central: bool = True
    ) -> "MovingEdgeCurrentView":
        """Create a new view filtered by position.

        Args:
            u: The u-coordinate.
            v: The v-coordinate.
            central: Whether to use central position.

        Returns:
            A new view with data filtered by the specified position.
        """
        rfs = at_position(self.rfs, u, v, central)
        currents = Namespace({
            cell_type: c[:, :, :, :, rfs[cell_type].index]
            for cell_type, c in self.currents.items()
        })
        return self.view(currents, rfs=rfs)

    def between_seconds(self, t_start: float, t_end: float) -> "MovingEdgeCurrentView":
        """Create a new view filtered by time range.

        Args:
            t_start: Start time in seconds.
            t_end: End time in seconds.

        Returns:
            A new view with data filtered by the specified time range.
        """
        slice = np.where((self.time >= t_start) & (self.time < t_end))[0]
        newview = self[:, :, slice, :]
        newview.time = self.time[slice]
        newview.responses = self.responses[:, :, slice]
        return newview

    def model_selection(self, mask: np.ndarray) -> "MovingEdgeCurrentView":
        """Create a new view with selected models.

        Args:
            mask: Boolean mask for model selection.

        Returns:
            A new view with selected models.
        """
        return self[mask, :, :, :]

    def __getattr__(self, key):
        return self.__getitem__(key)

    def __getitem__(self, key) -> Union["MovingEdgeCurrentView", Any]:
        # e.g. view.C3
        if isinstance(key, str) and key in self.source_types:
            return self.view(Namespace({key: self.currents[key]}))
        # e.g. view["C3", 0, 0, 0]
        elif (
            isinstance(key, Iterable)
            and isinstance(key[0], str)
            and key[0] in self.source_types
            and len(key[1:]) == self.shape
        ):
            return self.view(self.currents[key[0]][key[1:]])
        # e.g. view[index, :, :, :]
        elif isinstance(key, Iterable) and len(key) == len(self.shape):
            return self.view(
                Namespace({cell_type: c[key] for cell_type, c in self.currents.items()}),
                responses=self.responses[key[:-1]],
            )
        # view[:]
        elif key == slice(None):
            if len(self.currents) == 1:
                return next(iter(self.currents.values()))
            return self.currents
        return object.__getattribute__(self, key)

    def __repr__(self):
        cv = {ct: v.shape for ct, v in self.currents.items()}
        formatted_cv = ",\n        ".join(
            f"'{ct}': Array(shape={v})" for ct, v in cv.items()
        )
        return (
            f"{self.__class__.__name__}(\n"
            f"    ensemble={self.ensemble.name},\n"
            f"    target_type={self.target_type},\n"
            f"    currents={{\n        {formatted_cv}\n    }},\n"
            f"    rfs={self.rfs}\n"
            f")"
        )

    @property
    def shape(self):
        return next(iter(self.currents.values())).shape

    def sorting(self, average_over_models: bool = True, mode: str = "all") -> np.ndarray:
        """Sort cell types based on their contributions.

        Args:
            average_over_models: Whether to average over models.
            mode: Sorting mode ("all", "excitatory", or "inhibitory").

        Returns:
            Sorted array of cell types.

        Raises:
            ValueError: If an invalid mode is provided.
        """
        summed = self if len(self.shape) == 4 else self.sum_over_cells()
        signs = self.signs()
        if average_over_models:
            absmax = {
                k: v * signs[k]
                for k, v in valmap(
                    lambda v: np.nanmax(
                        np.abs(np.nanmean(v, axis=1, keepdims=True)),
                        axis=(0, 2, 3),
                    ),
                    summed[:],
                ).items()
            }
        else:
            # summing over on/off, angles and time to sort -- results in n_models sortings
            absmax = {
                k: v * signs[k]
                for k, v in valmap(
                    lambda v: np.nanmax(np.abs(v), axis=(0, 2, 3)), summed[:]
                ).items()
            }
        cell_types = np.array(list(absmax.keys()))
        values = np.array(list(absmax.values()))
        sorting = np.argsort(values, axis=0).T
        #         if average_over_models:
        #             # add extra dimension here for the next operation
        #             sorting = sorting[None]
        self.sorted_cell_types = cell_types[sorting[:, ::-1]]

        # return all excitatory and inhibitory from most excitatory to most inhibitory
        if mode == "all":
            return self.sorted_cell_types
        # from most excitatory to least excitatory
        elif mode == "excitatory":
            assert average_over_models
            return np.array([
                cell_type
                for cell_type in self.sorted_cell_types[0]
                if signs[cell_type] == 1
            ])
        # from most inhibitory to least inhibitory
        elif mode == "inhibitory":
            assert average_over_models
            return np.array([
                cell_type
                for cell_type in self.sorted_cell_types[0][::-1]
                if signs[cell_type] == -1
            ])
        else:
            raise ValueError(f"mode {mode}")

    def filter_cell_types_by_contribution(
        self,
        bins: int = 3,
        cut_off_edge: int = 1,
        mode: str = "above_cut_off",
        statistic: Callable = np.max,
    ) -> np.ndarray:
        """Filter cell types based on their contribution.

        Args:
            bins: Number of bins for contribution levels.
            cut_off_edge: Edge index for cut-off.
            mode: Filtering mode ("above_cut_off" or "below_cut_off").
            statistic: Function to compute the statistic.

        Returns:
            Filtered array of cell types.

        Raises:
            ValueError: If an invalid mode is provided.

        Info:
            In principle, chunks the y-axis of the current plots into excitatory and
            inhibitory parts and each of the parts into bins. All cell types with currents
            above or below, depending on the mode, the specified bin edge are discarded.
        """
        sorting = self.sorting()[0]
        signs = self.signs()
        currents = self.sum_over_cells().currents

        filtered_cell_types = []
        for sign in [1, -1]:
            # compute the std over all inputs
            values = {
                cell_type: statistic(np.abs(currents[cell_type][:]))
                for cell_type in sorting
                if signs[cell_type] == sign
            }
            # bin into three bins
            # ala (low contribution, medium contribution, high contribution)
            counts, bins = np.histogram(list(values.values()), bins=bins)
            cut_off_value = bins[cut_off_edge]
            if mode == "above_cut_off":
                filtered_cell_types.extend(
                    list(valfilter(lambda v, cut_off=cut_off_value: v >= cut_off, values))
                )
            elif mode == "below_cut_off":
                filtered_cell_types.extend(
                    list(valfilter(lambda v, cut_off=cut_off_value: v < cut_off, values))
                )
            else:
                raise ValueError(f"mode {mode}")
        return np.array(filtered_cell_types)

    def filter_source_types(
        self,
        hide_source_types: str | list | None,
        bins: int,
        edge: int,
        mode: str,
        statistic: Callable = np.max,
    ) -> np.ndarray:
        """Filter source types based on various criteria.

        Args:
            hide_source_types: Source types to hide or "auto".
            bins: Number of bins for contribution levels.
            edge: Edge index for cut-off.
            mode: Filtering mode.
            statistic: Function to compute the statistic.

        Returns:
            Filtered array of source types.
        """
        source_types = self.sorting()[0]
        if isinstance(hide_source_types, str) and hide_source_types == "auto":
            hide_source_types = self.filter_cell_types_by_contribution(
                bins=bins, cut_off_edge=edge, mode=mode, statistic=statistic
            )

        if hide_source_types is not None:
            source_types = np.array([
                source_type
                for source_type in source_types
                if source_type not in hide_source_types
            ])
        return source_types

    def signs(self) -> dict[str, float]:
        """Compute the signs of receptive fields for each source type.

        Returns:
            Dictionary of signs for each source type.
        """
        return {ct: np.mean(self.rfs[ct].sign) for ct in self.rfs.source_types}

    def sum_over_cells(self) -> "MovingEdgeCurrentView":
        """Sum currents over cells.

        Returns:
            A new view with currents summed over cells.
        """
        return self.view(
            Namespace({
                cell_type: c.sum(axis=-1) for cell_type, c in self.currents.items()
            }),
        )

    def plot_spatial_contribution(
        self,
        source_type: str,
        t_start: float,
        t_end: float,
        mode: str = "peak",
        title: str = "{source_type} :→",
        fig: plt.Figure | None = None,
        ax: plt.Axes | None = None,
        max_extent: float | None = None,
        **kwargs,
    ) -> plt.Axes:
        """Plot the spatial contribution of a source type.

        Args:
            source_type: The source type to plot.
            t_start: Start time for the plot.
            t_end: End time for the plot.
            mode: Mode for calculating values ("peak", "mean", or "std").
            title: Title format string for the plot.
            fig: Existing figure to use.
            ax: Existing axes to use.
            max_extent: Maximum extent of the spatial filter.
            **kwargs: Additional keyword arguments for plt_utils.kernel.

        Returns:
            Axes object containing the plot.
        """
        current_view = kwargs.get("current_view") or (
            self.between_seconds(t_start, t_end)  # .at_contrast(contrast).at_angle(angle)
        )

        vmin = kwargs.get("vmin") or (
            np.floor(
                min(
                    0,
                    min(
                        current.mean(axis=(0, 1, 2)).min()
                        for current in list(current_view[:].values())
                    ),
                )
                * 100
            )
            / 100
        )

        vmax = kwargs.get("vmax") or (
            np.ceil(
                max(
                    0,
                    max(
                        current.mean(axis=(0, 1, 2)).max()
                        for current in list(current_view[:].values())
                    ),
                )
                * 100
            )
            / 100
        )

        u, v = current_view.rfs[source_type][["source_u", "source_v"]].values.T
        # average over models
        # (1, n_models, 1, n_timesteps, n_models) -> (n_timesteps, n_models)
        # import pdb

        # pdb.set_trace()
        values = current_view[source_type][:].mean(axis=(0, 1))
        if mode == "peak":
            values = values[
                np.argmax(np.abs(values), axis=0), np.arange(values.shape[-1])
            ]
        elif mode == "mean":
            values = np.mean(values, axis=0)
        elif mode == "std":
            signs = self.signs()
            values = signs[source_type] * np.std(values, axis=0)
        fig, ax, _ = plots.kernel(
            u,
            v,
            values,
            fill=True,
            max_extent=max_extent or current_view.rfs.max_extent,
            label=title.format(source_type=source_type),
            labelxy="auto",
            strict_sign=False,
            fig=fig,
            ax=ax,
            **kwargs,
        )
        (xmin, ymin, xmax, ymax) = ax.dataLim.extents
        ax.set_xlim(plt_utils.get_lims((xmin, xmax), 0.01))
        ax.set_ylim(plt_utils.get_lims((ymin, ymax), 0.01))

    def plot_spatial_contribution_grid(
        self,
        t_start: float,
        t_end: float,
        max_extent: float = 3,
        mode: str = "peak",
        title: str = "{source_type} :→",
        fig: plt.Figure | None = None,
        axes: np.ndarray[plt.Axes] | None = None,
        fontsize: float = 5,
        edgewidth: float = 0.125,
        title_y: float = 0.8,
        max_figure_height_cm: float = 9.271,
        panel_height_cm: float | str = "auto",
        max_figure_width_cm: float = 2.54,
        panel_width_cm: float = 2.54,
        annotate: bool = False,
        cbar: bool = False,
        hide_source_types: str | list | None = "auto",
        hide_source_types_bins: int = 5,
        hide_source_types_cut_off_edge: int = 1,
        hide_source_types_mode: str = "below_cut_off",
        max_axes: int | None = None,
        **kwargs,
    ) -> tuple[
        plt.Figure,
        np.ndarray[plt.Axes],
        tuple[plt.Colorbar, plt.Colormap, plt.Normalize, float, float],
    ]:
        """Plot a grid of spatial contributions for different source types.

        Args:
            t_start: Start time for the plot.
            t_end: End time for the plot.
            max_extent: Maximum extent of the spatial filter.
            mode: Mode for calculating values ("peak", "mean", or "std").
            title: Title format string for each subplot.
            fig: Existing figure to use.
            axes: Existing axes to use.
            fontsize: Font size for labels and titles.
            edgewidth: Width of edges in the plot.
            title_y: Y-position of the title.
            max_figure_height_cm: Maximum figure height in centimeters.
            panel_height_cm: Height of each panel in centimeters.
            max_figure_width_cm: Maximum figure width in centimeters.
            panel_width_cm: Width of each panel in centimeters.
            annotate: Whether to annotate the plots.
            cbar: Whether to add a colorbar.
            hide_source_types: Source types to hide or "auto".
            hide_source_types_bins: Number of bins for auto-hiding.
            hide_source_types_cut_off_edge: Cut-off edge for auto-hiding.
            hide_source_types_mode: Mode for auto-hiding source types.
            max_axes: Maximum number of axes to create.
            **kwargs: Additional keyword arguments for plot_spatial_contribution.

        Returns:
            Figure, axes, and colorbar information (cbar, cmap, norm, vmin, vmax).
        """
        current_view = self.between_seconds(t_start, t_end)

        vmin = (
            np.floor(
                min(
                    0,
                    min(
                        current.mean(axis=(0, 1, 2)).min()
                        for current in list(current_view[:].values())
                    ),
                )
                * 10
            )
            / 10
        )

        vmax = (
            np.ceil(
                max(
                    0,
                    max(
                        current.mean(axis=(0, 1, 2)).max()
                        for current in list(current_view[:].values())
                    ),
                )
                * 10
            )
            / 10
        )

        source_types = self.filter_source_types(
            hide_source_types,
            bins=hide_source_types_bins,
            edge=hide_source_types_cut_off_edge,
            mode=hide_source_types_mode,
        )

        if fig is None and axes is None:
            figsize = figsize_from_n_items(
                max_axes or len(source_types),
                max_figure_height_cm=max_figure_height_cm,
                panel_height_cm=(
                    max_figure_height_cm / (max_axes or len(source_types))
                    if panel_height_cm == "auto"
                    else panel_height_cm
                ),
                max_figure_width_cm=max_figure_width_cm,
                panel_width_cm=panel_width_cm,
            )
            fig, axes = figsize.axis_grid(
                unmask_n=max_axes or len(source_types), hspace=0.1, wspace=0
            )
            if max_axes is not None and len(source_types) < max_axes:
                for ax in np.array(axes).flatten():
                    if isinstance(ax, Axes):
                        ax.axis("off")

        for i, source_type in enumerate(source_types):
            self.plot_spatial_contribution(
                source_type,
                #                 contrast,
                #                 angle,
                t_start,
                t_end,
                mode=mode,
                title=title,
                fontsize=fontsize,
                edgewidth=edgewidth,
                title_y=title_y,
                fig=fig,
                ax=axes[i],
                current_view=current_view,
                vmin=vmin,
                vmax=vmax,
                annotate=annotate,
                cbar=False,
                max_extent=max_extent or current_view.rfs.max_extent,
                **kwargs,
            )

        cmap = plt.cm.seismic
        norm = plt_utils.get_norm(vmin=vmin, vmax=vmax, midpoint=0)
        if cbar:
            cbar = plt_utils.add_colorbar_to_fig(
                fig,
                width=0.01,
                height=0.25,
                fontsize=fontsize,
                cmap=cmap,
                norm=norm,
                label=f"{mode} input currents",
                n_ticks=4,
                n_decimals=1,
            )
        return fig, axes, (cbar, cmap, norm, vmin, vmax)

    def plot_spatial_filter(
        self,
        source_type: str,
        title: str = "{source_type} :→",
        fig: plt.Figure | None = None,
        ax: plt.Axes | None = None,
        max_extent: float | None = None,
        **kwargs,
    ) -> plt.Axes:
        """Plot the spatial filter for a given source type.

        Args:
            source_type: The source type to plot.
            title: Title format string for the plot.
            fig: Existing figure to use.
            ax: Existing axes to use.
            max_extent: Maximum extent of the spatial filter.
            **kwargs: Additional keyword arguments for plt_utils.kernel.

        Returns:
            Axes object containing the plot.
        """
        filter = self.rfs

        def filter_values(rf):
            return (rf.n_syn * rf.sign).values

        vmin = kwargs.pop("vmin", None) or (
            np.floor(
                min(
                    0,
                    min(
                        min(filter_values(filter[source_type]))
                        for source_type in self.source_types
                    ),
                )
                * 100
            )
            / 100
        )

        vmax = kwargs.pop("vmax", None) or (
            np.ceil(
                max(
                    0,
                    max(
                        max(filter_values(filter[source_type]))
                        for source_type in self.source_types
                    ),
                )
                * 100
            )
            / 100
        )

        u, v = filter[source_type][["source_u", "source_v"]].values.T
        # average over models
        # (1, n_models, 1, n_timesteps, n_models) -> (n_timesteps, n_models)
        values = filter_values(filter[source_type])

        label = title.format(source_type=source_type)
        fig, ax, _ = plt_utils.kernel(
            u,
            v,
            values,
            fill=True,
            max_extent=max_extent or filter.max_extent,
            label=label,
            labelxy="auto",
            strict_sign=False,
            fig=fig,
            ax=ax,
            vmin=vmin,
            vmax=vmax,
            **kwargs,
        )
        (xmin, ymin, xmax, ymax) = ax.dataLim.extents
        ax.set_xlim(plt_utils.get_lims((xmin, xmax), 0.01))
        ax.set_ylim(plt_utils.get_lims((ymin, ymax), 0.01))
        return ax

    def plot_spatial_filter_grid(
        self,
        title: str = "{source_type} :→",
        fig: plt.Figure | None = None,
        axes: np.ndarray[plt.Axes] | None = None,
        max_extent: float | None = None,
        fontsize: float = 5,
        edgewidth: float = 0.125,
        title_y: float = 0.8,
        max_figure_height_cm: float = 9.271,
        panel_height_cm: float | str = "auto",
        max_figure_width_cm: float = 2.54,
        panel_width_cm: float = 2.54,
        annotate: bool = False,
        cbar: bool = False,
        hide_source_types: str | list | None = "auto",
        hide_source_types_bins: int = 5,
        hide_source_types_cut_off_edge: int = 1,
        hide_source_types_mode: str = "below_cut_off",
        max_axes: int | None = None,
        wspace: float = 0.0,
        hspace: float = 0.1,
        **kwargs,
    ) -> tuple[
        plt.Figure,
        np.ndarray[plt.Axes],
        tuple[plt.Colorbar, plt.Colormap, plt.Normalize, float, float],
    ]:
        """Plot a grid of spatial filters for different source types.

        Args:
            title: Title format string for each subplot.
            fig: Existing figure to use.
            axes: Existing axes to use.
            max_extent: Maximum extent of the spatial filter.
            fontsize: Font size for labels and titles.
            edgewidth: Width of edges in the plot.
            title_y: Y-position of the title.
            max_figure_height_cm: Maximum figure height in centimeters.
            panel_height_cm: Height of each panel in centimeters.
            max_figure_width_cm: Maximum figure width in centimeters.
            panel_width_cm: Width of each panel in centimeters.
            annotate: Whether to annotate the plots.
            cbar: Whether to add a colorbar.
            hide_source_types: Source types to hide or "auto".
            hide_source_types_bins: Number of bins for auto-hiding.
            hide_source_types_cut_off_edge: Cut-off edge for auto-hiding.
            hide_source_types_mode: Mode for auto-hiding source types.
            max_axes: Maximum number of axes to create.
            wspace: Width space between subplots.
            hspace: Height space between subplots.
            **kwargs: Additional keyword arguments for plot_spatial_filter.

        Returns:
            Figure, axes, and colorbar information (cbar, cmap, norm, vmin, vmax).
        """
        filter = self.rfs

        def filter_values(rf):
            return (rf.n_syn * rf.sign).values

        vmin = kwargs.pop("vmin", None) or (
            np.floor(
                min(
                    0,
                    min(
                        min(filter_values(filter[source_type]))
                        for source_type in self.source_types
                    ),
                )
                * 100
            )
            / 100
        )

        vmax = kwargs.pop("vmax", None) or (
            np.ceil(
                max(
                    0,
                    max(
                        max(filter_values(filter[source_type]))
                        for source_type in self.source_types
                    ),
                )
                * 100
            )
            / 100
        )

        source_types = self.filter_source_types(
            hide_source_types,
            bins=hide_source_types_bins,
            edge=hide_source_types_cut_off_edge,
            mode=hide_source_types_mode,
        )

        if fig is None and axes is None:
            figsize = figsize_from_n_items(
                max_axes or len(source_types),
                max_figure_height_cm=max_figure_height_cm,
                panel_height_cm=(
                    max_figure_height_cm / (max_axes or len(source_types))
                    if panel_height_cm == "auto"
                    else panel_height_cm
                ),
                max_figure_width_cm=max_figure_width_cm,
                panel_width_cm=panel_width_cm,
            )
            fig, axes = figsize.axis_grid(
                unmask_n=max_axes or len(source_types), hspace=hspace, wspace=wspace
            )
            if max_axes is not None and len(source_types) < max_axes:
                for ax in np.array(axes).flatten():
                    if isinstance(ax, Axes):
                        ax.axis("off")

        for i, source_type in enumerate(source_types):
            self.plot_spatial_filter(
                source_type,
                title=title,
                fontsize=fontsize,
                edgewidth=edgewidth,
                title_y=title_y,
                fig=fig,
                ax=axes[i],
                vmin=vmin,
                vmax=vmax,
                annotate=annotate,
                cbar=False,
                max_extent=max_extent or filter.max_extent,
                **kwargs,
            )

        cmap = plt.cm.seismic
        norm = plt_utils.get_norm(vmin=vmin, vmax=vmax, midpoint=0)
        if cbar:
            cbar = plt_utils.add_colorbar_to_fig(
                fig,
                width=0.01,
                height=0.25,
                fontsize=fontsize,
                cmap=cmap,
                norm=norm,
                label="spatial filters",
                n_ticks=4,
                n_decimals=1,
            )
        return fig, axes, (cbar, cmap, norm, vmin, vmax)

    def view(
        self,
        currents: Namespace,
        rfs: ReceptiveFields | None = None,
        time: np.ndarray | None = None,
        responses: np.ndarray | None = None,
        arg_df: pd.DataFrame | None = None,
    ) -> "MovingEdgeCurrentView":
        """
        Create a new view with the given currents, rfs, time, responses, and arg_df.

        Args:
            currents: Currents for each source type.
            rfs: Receptive fields for the target cells.
            time: Time array for the simulation.
            responses: Responses of the target cells.
            arg_df: DataFrame containing stimulus arguments.

        Returns:
            A new view with the given data.
        """
        arg_df = arg_df.reset_index(drop=True) if arg_df is not None else self.arg_df
        return MovingEdgeCurrentView(
            self.ensemble,
            self.target_type,
            self.exp_data,
            arg_df,
            currents,
            rfs if rfs is not None else self.rfs,
            time if time is not None else self.time,
            responses if responses is not None else self.responses,
        )

    def subtract_baseline(self) -> "MovingEdgeCurrentView":
        """
        Create a new view with baseline subtracted from the currents and responses.

        Returns:
            A new view with baseline subtracted data.
        """
        return self.view(
            Namespace({
                cell_type: c - np.take(c, [0], -2)
                for cell_type, c in self.currents.items()
            }),
            responses=self.responses - np.take(self.responses, [0], -1),
        )

    def subtract_mean(self) -> "MovingEdgeCurrentView":
        """
        Create a new view with mean subtracted from the currents and responses.

        Returns:
            A new view with mean subtracted data.
        """
        return self.view(
            Namespace({
                cell_type: c - np.mean(c, -2, keepdims=True)
                for cell_type, c in self.currents.items()
            }),
            responses=self.responses - np.mean(self.responses, -1, keepdims=True),
        )

    def standardize(self) -> "MovingEdgeCurrentView":
        """
        Create a new view with standardized currents and responses.

        Returns:
            A new view with standardized data.
        """
        return self.view(
            Namespace({
                cell_type: (c - np.mean(c, -2, keepdims=True))
                / (np.std(c, -2, keepdims=True) + 1e-15)
                for cell_type, c in self.currents.items()
            }),
            responses=(self.responses - np.mean(self.responses, -1, keepdims=True))
            / (np.std(self.responses, -1, keepdims=True) + 1e-15),
        )

    def standardize_over_time_and_pd_nd(
        self, t_start: float, t_end: float, pd: float
    ) -> "MovingEdgeCurrentView":
        """
        Create a new view with standardized currents and responses over time and PD/ND.

        Args:
            t_start: Start time for standardization.
            t_end: End time for standardization.
            pd: Preferred direction for standardization.

        Returns:
            A new view with standardized data.
        """
        temp = self.between_seconds(t_start, t_end).at_angle([pd, (pd - 180) % 360])
        return self.view(
            Namespace({
                cell_type: (
                    c - np.mean(temp.currents[cell_type], (-2, -3), keepdims=True)
                )
                / (np.std(temp.currents[cell_type], (-2, -3), keepdims=True) + 1e-15)
                for cell_type, c in self.currents.items()
            }),
            responses=(self.responses - np.mean(temp.responses, (-1, -2), keepdims=True))
            / (np.std(temp.responses, (-1, -2), keepdims=True) + 1e-15),
        )

    def init_colors(self, source_types: list[str]) -> None:
        """
        Initialize colors for source types.

        Args:
            source_types: List of source types.
        """
        signs = self.signs()
        signs = {cell_type: signs[cell_type] for cell_type in source_types}
        signs_reversed = {cell_type: signs[cell_type] for cell_type in source_types[::-1]}
        n_exc = len([v for v in signs.values() if v == 1])
        n_inh = len([v for v in signs.values() if v == -1])
        exc_colors_pd = cmap_iter(
            truncate_colormap(plt.cm.RdBu, minval=0.05, maxval=0.45, n=n_exc)
        )
        inh_cmap_pd = cmap_iter(
            truncate_colormap(plt.cm.RdBu_r, minval=0.05, maxval=0.45, n=n_inh)
        )
        exc_colors_nd = cmap_iter(
            truncate_colormap(plt.cm.BrBG_r, minval=0.05, maxval=0.45, n=n_exc)
        )
        inh_cmap_nd = cmap_iter(
            truncate_colormap(plt.cm.BrBG, minval=0.05, maxval=0.45, n=n_inh)
        )
        colors_pd = {}
        colors_nd = {}
        for _, (cell_type, sign) in enumerate(signs.items()):
            if sign == 1:
                # take the first half of the RdBu colormap, i.e. red
                colors_pd[cell_type] = next(exc_colors_pd)
                colors_nd[cell_type] = next(exc_colors_nd)

        for _, (cell_type, sign) in enumerate(signs_reversed.items()):
            if sign == -1:
                # take the second half of the RdBu colormap, i.e. blue
                colors_pd[cell_type] = next(inh_cmap_pd)
                colors_nd[cell_type] = next(inh_cmap_nd)
        self.colors_pd = colors_pd
        self.colors_nd = colors_nd

    def color(self, source_type: str, pd: bool = True) -> tuple[float, float, float]:
        """
        Get the color for a given source type.

        Args:
            source_type: The source type.
            pd: Whether to use PD or ND colors.

        Returns:
            The color as an RGB tuple.
        """
        if pd:
            return self.colors_pd[source_type]
        return self.colors_nd[source_type]

    def zorder(
        self,
        source_types: list[str],
        source_type: str,
        start_exc: int = 1000,
        start_inh: int = 1000,
    ) -> int:
        """
        Get the z-order for a given source type.

        Args:
            source_types: List of source types.
            source_type: The source type.
            start_exc: Starting z-order for excitatory cells.
            start_inh: Starting z-order for inhibitory cells.

        Returns:
            The z-order for the given source type.
        """
        signs = self.signs()
        signs_reversed = {cell_type: signs[cell_type] for cell_type in source_types[::-1]}

        z_order = start_exc
        for _, (cell_type, sign) in enumerate(signs.items()):
            if sign == 1:
                if cell_type == source_type:
                    return z_order
                z_order -= 10

        z_order = start_inh
        for _, (cell_type, sign) in enumerate(signs_reversed.items()):
            if sign == -1:
                if cell_type == source_type:
                    return z_order
                z_order -= 10

    def ylims(
        self, source_types: list[str] | None = None, offset: float = 0.02
    ) -> dict[str, tuple[float, float]]:
        """
        Get the y-limits for temporal contributions summed over cells.

        Args:
            source_types: List of source types to consider.
            offset: Offset for the y-limits.

        Returns:
            Y-limits for the given source types or all source types.
        """
        if source_types is not None:
            return {
                cell_type: plt_utils.get_lims(c, offset)
                for cell_type, c in self.sum_over_cells().currents.items()
                if cell_type in source_types
            }
        return plt_utils.get_lims(list(self.sum_over_cells().currents.values()), offset)

    def plot_response(
        self,
        contrast: float,
        angle: float,
        t_start: float = 0,
        t_end: float = 1,
        max_figure_height_cm: float = 1.4477,
        panel_height_cm: float = 1.4477,
        max_figure_width_cm: float = 4.0513,
        panel_width_cm: float = 4.0513,
        fontsize: float = 5,
        model_average: bool = True,
        color: tuple[float, float, float] = (0, 0, 0),
        legend: bool = False,
        hide_yaxis: bool = True,
        trim_axes: bool = True,
        quantile: float | None = None,
        scale_position: str | None = None,  # "lower left",
        scale_label: str = "{:.0f} ms",
        scale_unit: float = 1000,
        hline: bool = False,
        fig: plt.Figure | None = None,
        ax: plt.Axes | None = None,
    ):
        """
        Plot the response to a moving edge stimulus.

        Args:
            contrast: The contrast of the stimulus.
            angle: The angle of the stimulus.
            t_start: Start time for the plot.
            t_end: End time for the plot.
            max_figure_height_cm: Maximum figure height in centimeters.
            panel_height_cm: Height of each panel in centimeters.
            max_figure_width_cm: Maximum figure width in centimeters.
            panel_width_cm: Width of each panel in centimeters.
            fontsize: Font size for labels and titles.
            model_average: Whether to plot the model average.
            color: Color for the plot.
            legend: Whether to show the legend.
            hide_yaxis: Whether to hide the y-axis.
            trim_axes: Whether to trim the axes.
            quantile: Quantile for shading.
            scale_position: Position of the scale.
            scale_label: Label format for the scale.
            scale_unit: Unit for the scale.
            hline: Whether to show a horizontal line at 0.
            fig: Existing figure to use.
            ax: Existing axes to use.

        Returns:
            Figure and axes objects.
        """
        r_pd = (
            self.at_angle(angle)
            .at_contrast(contrast)
            .between_seconds(t_start, t_end)
            .responses.squeeze(axis=-2)
        )
        r_nd = (
            self.at_angle((angle - 180) % 360)
            .at_contrast(contrast)
            .between_seconds(t_start, t_end)
            .responses.squeeze(axis=-2)
        )

        if fig is None and ax is None:
            figsize = figsize_from_n_items(
                1,
                max_figure_height_cm=max_figure_height_cm,
                panel_height_cm=panel_height_cm,
                max_figure_width_cm=max_figure_width_cm,
                panel_width_cm=panel_width_cm,
            )
            fig, axes = figsize.axis_grid(hspace=0.0, wspace=0, fontsize=fontsize)
            ax = axes[0]

        color = [hex2color(PD), hex2color(ND)] if color is None else [color, color]

        if model_average:
            fig, ax, _, _ = plots.traces(
                [r_pd.mean(axis=0), r_nd.mean(axis=0)],
                x=self.between_seconds(t_start, t_end).time,
                color=color,
                linewidth=1,
                fontsize=fontsize,
                null_line=False,
                fig=fig,
                ax=ax,
                linestyle=["solid", "dashed"],
                legend="" if not legend else [f"{self.target_type}", "null direction"],
                scale_pos=scale_position,
                scale_label=scale_label,
                scale_unit=scale_unit,
            )
        else:
            fig, ax, _, _ = plots.traces(
                r_pd,
                x=self.between_seconds(t_start, t_end).time,
                mean_color=adapt_color_alpha(color[0], 1),
                color=adapt_color_alpha(color[0], 0.5),
                linewidth=0.25,
                zorder_traces=0,
                zorder_mean=10,
                fontsize=fontsize,
                null_line=False,
                highlight_mean=True,
                fig=fig,
                ax=ax,
            )
            plots.traces(
                r_nd,
                x=self.between_seconds(t_start, t_end).time,
                mean_color=adapt_color_alpha(color[1], 1),
                color=adapt_color_alpha(color[1], 0.5),
                linewidth=0.25,
                zorder_traces=0,
                zorder_mean=10,
                fontsize=fontsize,
                null_line=False,
                highlight_mean=True,
                fig=fig,
                linestyle="dashed",
                ax=ax,
            )
        if quantile:
            quantile_pd = np.quantile(r_pd, quantile, axis=0)
            quantile_nd = np.quantile(r_nd, quantile, axis=0)
            ax.fill_between(
                self.between_seconds(t_start, t_end).time,
                quantile_pd[0],
                quantile_pd[1],
                facecolor=adapt_color_alpha(color[0], 0.1),
                edgecolor=adapt_color_alpha(color[0], 1),
                linewidth=0.25,
            )
            ax.fill_between(
                self.between_seconds(t_start, t_end).time,
                quantile_nd[0],
                quantile_nd[1],
                facecolor=adapt_color_alpha(color[1], 0.1),
                edgecolor=adapt_color_alpha(color[1], 1),
                linewidth=0.25,
                linestyle="dashed",
            )

        if hline:
            # horizontal line at 0
            ax.axhline(0, color=(0, 0, 0, 1), linewidth=0.25, zorder=-10)

        if hide_yaxis:
            plt_utils.rm_spines(ax, ("left",))
        if trim_axes:
            plt_utils.trim_axis(ax)
        if legend:
            ax.legend(
                fontsize=fontsize,
                ncols=1,
                bbox_to_anchor=(1.05, 1),
                loc="upper left",
                borderaxespad=0.0,
            )
        return fig, ax

    def plot_response_pc_nc(
        self,
        contrast: float,
        angle: float,
        t_start: float = 0,
        t_end: float = 1,
        max_figure_height_cm: float = 1.4477,
        panel_height_cm: float = 1.4477,
        max_figure_width_cm: float = 4.0513,
        panel_width_cm: float = 4.0513,
        fontsize: float = 5,
        model_average: bool = True,
        color: tuple[float, float, float] = (0, 0, 0),
        legend: bool = False,
        hide_yaxis: bool = True,
        trim_axes: bool = True,
        quantile: float | None = None,
        scale_position: str | None = None,
        scale_label: str = "{:.0f} ms",
        scale_unit: float = 1000,
        fig: plt.Figure | None = None,
        ax: plt.Axes | None = None,
        hline: bool = False,
    ) -> tuple[plt.Figure, plt.Axes]:
        """
        Plot the response to a moving edge stimulus with positive and negative contrasts.

        Args:
            contrast: The contrast of the stimulus.
            angle: The angle of the stimulus.
            t_start: Start time for the plot.
            t_end: End time for the plot.
            max_figure_height_cm: Maximum figure height in centimeters.
            panel_height_cm: Height of each panel in centimeters.
            max_figure_width_cm: Maximum figure width in centimeters.
            panel_width_cm: Width of each panel in centimeters.
            fontsize: Font size for labels and titles.
            model_average: Whether to plot the model average.
            color: Color for the plot.
            legend: Whether to show the legend.
            hide_yaxis: Whether to hide the y-axis.
            trim_axes: Whether to trim the axes.
            quantile: Quantile for shading.
            scale_position: Position of the scale.
            scale_label: Label format for the scale.
            scale_unit: Unit for the scale.
            fig: Existing figure to use.
            ax: Existing axes to use.
            hline: Whether to show a horizontal line at 0.

        Returns:
            Figure and axes objects.
        """
        r_pc = (
            self.at_angle(angle)
            .at_contrast(contrast)
            .between_seconds(t_start, t_end)
            .responses.squeeze(axis=-2)
        )
        r_nc = (
            self.at_angle(angle)
            .at_contrast(0 if contrast == 1 else 1)
            .between_seconds(t_start, t_end)
            .responses.squeeze(axis=-2)
        )

        if fig is None and ax is None:
            figsize = figsize_from_n_items(
                1,
                max_figure_height_cm=max_figure_height_cm,
                panel_height_cm=panel_height_cm,
                max_figure_width_cm=max_figure_width_cm,
                panel_width_cm=panel_width_cm,
            )
            fig, axes = figsize.axis_grid(hspace=0.0, wspace=0, fontsize=fontsize)
            ax = axes[0]

        color = [hex2color(PD), hex2color(ND)] if color is None else [color, color]

        if model_average:
            fig, ax, _, _ = plots.traces(
                [r_pc.mean(axis=0), r_nc.mean(axis=0)],
                x=self.between_seconds(t_start, t_end).time,
                color=color,
                linewidth=1,
                fontsize=fontsize,
                null_line=False,
                fig=fig,
                ax=ax,
                linestyle=["solid", "dotted"],
                legend="" if not legend else [f"{self.target_type}", "null contrast"],
                scale_pos=scale_position,
                scale_label=scale_label,
                scale_unit=scale_unit,
            )
        else:
            fig, ax, _, _ = plots.traces(
                r_pc,
                x=self.between_seconds(t_start, t_end).time,
                mean_color=adapt_color_alpha(color[0], 1),
                color=adapt_color_alpha(color[0], 0.5),
                linewidth=0.25,
                zorder_traces=0,
                zorder_mean=10,
                fontsize=fontsize,
                null_line=False,
                highlight_mean=True,
                fig=fig,
                ax=ax,
            )
            plots.traces(
                r_nc,
                x=self.between_seconds(t_start, t_end).time,
                mean_color=adapt_color_alpha(color[1], 1),
                color=adapt_color_alpha(color[1], 0.5),
                linewidth=0.25,
                zorder_traces=0,
                zorder_mean=10,
                fontsize=fontsize,
                null_line=False,
                highlight_mean=True,
                fig=fig,
                linestyle="dashed",
                ax=ax,
            )
        if quantile:
            quantile_pd = np.quantile(r_pc, quantile, axis=0)
            quantile_nd = np.quantile(r_nc, quantile, axis=0)
            ax.fill_between(
                self.between_seconds(t_start, t_end).time,
                quantile_pd[0],
                quantile_pd[1],
                facecolor=adapt_color_alpha(color[0], 0.1),
                edgecolor=adapt_color_alpha(color[0], 1),
                linewidth=0.25,
            )
            ax.fill_between(
                self.between_seconds(t_start, t_end).time,
                quantile_nd[0],
                quantile_nd[1],
                facecolor=adapt_color_alpha(color[1], 0.1),
                edgecolor=adapt_color_alpha(color[1], 1),
                linewidth=0.25,
                linestyle="dashed",
            )

        # horizontal line at 0
        if hline:
            ax.axhline(0, color=(0, 0, 0, 1), linewidth=0.25, zorder=-10)

        if hide_yaxis:
            plt_utils.rm_spines(ax, ("left",))
        if trim_axes:
            plt_utils.trim_axis(ax)
        if legend:
            ax.legend(
                fontsize=fontsize,
                ncols=1,
                bbox_to_anchor=(1.05, 1),
                loc="upper left",
                borderaxespad=0.0,
            )
        return fig, ax

    def plot_temporal_contributions(
        self,
        contrast: float,
        angle: float,
        t_start: float = 0,
        t_end: float = 1,
        fontsize: float = 5,
        linewidth: float = 0.25,
        legend: bool = False,
        legend_standalone: bool = True,
        legend_figsize_cm: tuple[float, float] = (4.0572, 1),
        legend_n_rows: int | None = None,
        max_figure_height_cm: float = 3.3941,
        panel_height_cm: float = 3.3941,
        max_figure_width_cm: float = 4.0572,
        panel_width_cm: float = 4.0572,
        model_average: bool = True,
        highlight_mean: bool = True,  # only applies if model_average is False
        sum_exc_inh: bool = False,
        only_sum: bool = False,
        hide_source_types: str | list | None = "auto",
        hide_source_types_bins: int = 5,
        hide_source_types_cut_off_edge: int = 1,
        hide_source_types_mode: str = "below_cut_off",
        hide_yaxis: bool = True,
        trim_axes: bool = True,
        quantile: float | None = None,
        fig: plt.Figure | None = None,
        ax: plt.Axes | None = None,
        legend_ax: plt.Axes | None = None,
        hline: bool = True,
        legend_n_cols: int | None = None,
        baseline_color: tuple[float, float, float, float] | None = None,
        colors: dict[str, tuple[float, float, float, float]] | None = None,
    ):
        """
        Plot temporal contributions of different source types.

        Args:
            contrast: The contrast of the stimulus.
            angle: The angle of the stimulus.
            t_start: Start time for the plot.
            t_end: End time for the plot.
            fontsize: Font size for labels and titles.
            linewidth: Line width for traces.
            legend: Whether to show the legend.
            legend_standalone: Whether to create a standalone legend.
            legend_figsize_cm: Figure size for the standalone legend.
            legend_n_rows: Number of rows for the standalone legend.
            max_figure_height_cm: Maximum figure height in centimeters.
            panel_height_cm: Height of each panel in centimeters.
            max_figure_width_cm: Maximum figure width in centimeters.
            panel_width_cm: Width of each panel in centimeters.
            model_average: Whether to plot the model average.
            highlight_mean: Whether to highlight the mean trace.
            sum_exc_inh: Whether to sum excitatory and inhibitory contributions.
            only_sum: Whether to only plot the summed contributions.
            hide_source_types: Source types to hide or "auto".
            hide_source_types_bins: Number of bins for auto-hiding.
            hide_source_types_cut_off_edge: Cut-off edge for auto-hiding.
            hide_source_types_mode: Mode for auto-hiding source types.
            hide_yaxis: Whether to hide the y-axis.
            trim_axes: Whether to trim the axes.
            quantile: Quantile for shading.
            fig: Existing figure to use.
            ax: Existing axes to use.
            legend_ax: Existing axes for the standalone legend.
            hline: Whether to show a horizontal line at 0.
            legend_n_cols: Number of columns for the standalone legend.
            baseline_color: Color for the baseline.
            colors: Colors for each source type.

        Returns:
            Figure, axes, and legend axes objects.

        Example:
            ```
            view = MovingEdgeCurrentView(...)
            fig, ax = view.plot_temporal_contributions(
                contrast=1.0,
                angle=0,
                t_start=0,
                t_end=1,
                fontsize=5,
                linewidth=0.25,
                legend=True
            )
            ```
        """
        if fig is None and ax is None:
            figsize = figsize_from_n_items(
                1,
                max_figure_height_cm=max_figure_height_cm,
                panel_height_cm=panel_height_cm,
                max_figure_width_cm=max_figure_width_cm,
                panel_width_cm=panel_width_cm,
            )
            fig, axes = figsize.axis_grid(hspace=0.0, wspace=0, fontsize=fontsize)
            ax = axes[0]
        cv_pd = (
            self.at_contrast(contrast)
            .at_angle(angle)
            .between_seconds(t_start, t_end)
            .sum_over_cells()
        )
        cv_nd = (
            self.at_contrast(contrast)
            .at_angle((angle - 180) % 360)
            .between_seconds(t_start, t_end)
            .sum_over_cells()
        )

        source_types = (
            self.at_contrast(contrast)
            .at_angle([angle, (angle - 180) % 360])
            .between_seconds(t_start, t_end)
            .filter_source_types(
                hide_source_types,
                hide_source_types_bins,
                hide_source_types_cut_off_edge,
                hide_source_types_mode,
            )
        )

        color_source_types = (
            self.at_contrast(contrast)
            .at_angle([angle, (angle - 180) % 360])
            .between_seconds(t_start, t_end)
            .filter_source_types(
                None,
                hide_source_types_bins,
                hide_source_types_cut_off_edge,
                hide_source_types_mode,
            )
        )
        cv_pd.init_colors(color_source_types)
        cv_nd.init_colors(color_source_types)

        def plot_mean_trace(
            time, trace, label, color, zorder, linestyle="solid", ax=None, fig=None
        ):
            ax.plot(
                time,
                trace,
                label=label,
                color=color,
                zorder=zorder,
                linestyle=linestyle,
            )

        def plot_individual_traces(
            traces, time, color, zorder, label, linestyle="solid", legend=None
        ):
            if not only_sum and not model_average:
                plots.traces(
                    traces,
                    time,
                    mean_color=color,
                    color=color,
                    linewidth=linewidth,
                    zorder_traces=0,
                    zorder_mean=zorder,
                    fontsize=fontsize,
                    null_line=True,
                    highlight_mean=highlight_mean,
                    fig=fig,
                    ax=ax,
                    legend=legend or label,
                    linestyle=linestyle,
                )

        def plot_quantile(traces, time, color, zorder, linestyle="solid"):
            if quantile:
                Q = np.quantile(traces, quantile, axis=0)
                ax.fill_between(
                    time,
                    Q[0],
                    Q[1],
                    facecolor=adapt_color_alpha(color, 0.1),
                    edgecolor=color,
                    linewidth=0.25,
                    linestyle=linestyle,
                    zorder=zorder - 1,
                )

        def plot_summed_trace(time, trace, label, color, zorder, linestyle="solid"):
            if np.any(trace):
                ax.plot(
                    time,
                    trace,
                    label=label,
                    color=color,
                    zorder=zorder,
                    linestyle=linestyle,
                )

        def get_summed_traces(signs, source_types, cv_pd, cv_nd):
            # sum over cell types then average over models
            exc_pd = np.zeros(cv_pd.shape)
            inh_pd = np.zeros(cv_pd.shape)
            exc_nd = np.zeros(cv_nd.shape)
            inh_nd = np.zeros(cv_nd.shape)
            # sum over cell types
            for source_type in source_types:
                if signs[source_type] == 1:
                    exc_pd += cv_pd[source_type][:]  # (1, n_models, 1, n_timesteps)
                    exc_nd += cv_nd[source_type][:]
                else:
                    inh_pd += cv_pd[source_type][:]
                    inh_nd += cv_nd[source_type][:]
            # (n_models, n_timesteps)
            return (
                exc_pd.squeeze(),
                inh_pd.squeeze(),
                exc_nd.squeeze(),
                inh_nd.squeeze(),
            )

        for source_type in source_types:
            if model_average and not only_sum:
                # mean traces solid for PD and dashed for ND
                if baseline_color is not None:
                    color = baseline_color
                elif colors:
                    color = colors[source_type]
                else:
                    color = cv_pd.color(source_type)

                plot_mean_trace(
                    cv_pd.time,
                    cv_pd[source_type][:].squeeze(axis=-2).T.mean(axis=1),
                    source_type,
                    color,
                    cv_pd.zorder(source_types, source_type),
                    ax=ax,
                    fig=fig,
                )
                plot_mean_trace(
                    cv_nd.time,
                    cv_nd[source_type][:].squeeze(axis=-2).T.mean(axis=1),
                    source_type,
                    color,
                    linestyle="dashed",
                    zorder=cv_pd.zorder(source_types, source_type),
                    ax=ax,
                    fig=fig,
                )

            elif not model_average and not only_sum:
                # individual traces
                plot_individual_traces(
                    cv_pd[source_type][:].squeeze(axis=-2),
                    cv_pd.time,
                    cv_pd.color(source_type),
                    cv_pd.zorder(source_types, source_type),
                    source_type,
                )
                plot_individual_traces(
                    cv_nd[source_type][:].squeeze(axis=-2),
                    cv_nd.time,
                    cv_pd.color(source_type),
                    cv_pd.zorder(source_types, source_type),
                    source_type,
                    linestyle="dashed",
                    legend="null direction",
                )

            # quantiles
            plot_quantile(
                cv_pd[source_type][:].squeeze(axis=-2),
                cv_pd.time,
                cv_pd.color(source_type),
                cv_pd.zorder(source_types, source_type),
                linestyle="solid",
            )
            plot_quantile(
                cv_nd[source_type][:].squeeze(axis=-2),
                cv_nd.time,
                cv_pd.color(source_type),
                cv_pd.zorder(source_types, source_type),
                linestyle="dashed",
            )
        if sum_exc_inh or only_sum:
            # plot summed traces
            signs = cv_pd.signs()
            exc_pd, inh_pd, exc_nd, inh_nd = get_summed_traces(
                signs, source_types, cv_pd, cv_nd
            )
            plot_summed_trace(
                cv_pd.time,
                exc_pd.mean(axis=0),
                "excitatory",
                (0.931, 0.0, 0.0, 1.0),
                zorder=2000,
            )
            plot_quantile(
                exc_pd,
                cv_pd.time,
                (0.931, 0.0, 0.0, 1.0),
                zorder=0,
                linestyle="solid",
            )
            plot_summed_trace(
                cv_nd.time,
                exc_nd.mean(axis=0),
                "excitatory",
                (0.931, 0.0, 0.0, 1.0),
                zorder=2000,
                linestyle="dashed",
            )
            plot_quantile(
                exc_nd,
                cv_pd.time,
                (0.931, 0.0, 0.0, 1.0),
                zorder=0,
                linestyle="dashed",
            )
            plot_summed_trace(
                cv_pd.time,
                inh_pd.mean(axis=0),
                "inhibitory",
                (0.0, 0.0, 0.849, 1.0),
                zorder=2000,
            )
            plot_quantile(
                inh_pd,
                cv_pd.time,
                (0.0, 0.0, 0.849, 1.0),
                zorder=0,
                linestyle="solid",
            )
            plot_summed_trace(
                cv_nd.time,
                inh_nd.mean(axis=0),
                "inhibitory",
                (0.0, 0.0, 0.849, 1.0),
                zorder=2000,
                linestyle="dashed",
            )
            plot_quantile(
                inh_nd,
                cv_pd.time,
                (0.0, 0.0, 0.849, 1.0),
                zorder=0,
                linestyle="dashed",
            )

        if hline:
            ax.hlines(
                0,
                cv_pd.time.min(),
                cv_pd.time.max(),
                color=(0, 0, 0, 1),
                linewidth=0.25,
                zorder=-10,
            )

        if legend:
            ax.legend(
                fontsize=fontsize,
                ncols=1,
                bbox_to_anchor=(1.05, 1),
                loc="upper left",
                borderaxespad=0.0,
            )
        else:
            ax.legend().set_visible(False)

        ax.set_xlabel("time (s)", fontsize=fontsize)
        #         ax.set_ylabel("current (a.u.)", fontsize=fontsize)

        if hide_yaxis:
            plt_utils.rm_spines(ax, ("left",))

        if trim_axes:
            plt_utils.trim_axis(ax)

        if legend_standalone:
            handles, labels = ax.get_legend_handles_labels()
            nd_handle = Line2D(
                [0], [0], color="k", lw=1, label="null direction", ls="dashed"
            )
            legend_n_rows = legend_n_rows or len(labels) + 1
            # legend_n_cols = (len(labels) + 1) // legend_n_rows
            legend_fig, legend_ax = plt_utils.standalone_legend(
                [*labels[::2], "null direction"],
                None,
                [*handles[::2], nd_handle],
                fontsize=fontsize,
                n_cols=legend_n_cols,
                handlelength=2,
                columnspacing=0.8,
                labelspacing=0.25,
                figsize=cm_to_inch(legend_figsize_cm),
                fig=fig if legend_ax is not None else None,
                ax=legend_ax,
            )
            return fig, ax, legend_fig, legend_ax
        return fig, ax

    def plot_temporal_contributions_pc_nc(
        self,
        contrast: float,
        angle: float,
        t_start: float = 0,
        t_end: float = 1,
        fontsize: float = 5,
        linewidth: float = 0.25,
        legend: bool = False,
        legend_standalone: bool = True,
        legend_figsize_cm: tuple[float, float] = (4.0572, 1),
        legend_n_rows: int | None = None,
        max_figure_height_cm: float = 3.3941,
        panel_height_cm: float = 3.3941,
        max_figure_width_cm: float = 4.0572,
        panel_width_cm: float = 4.0572,
        model_average: bool = True,
        highlight_mean: bool = True,
        sum_exc_inh: bool = False,
        only_sum: bool = False,
        hide_source_types: str | list | None = "auto",
        hide_source_types_bins: int = 5,
        hide_source_types_cut_off_edge: int = 1,
        hide_source_types_mode: str = "below_cut_off",
        hide_yaxis: bool = True,
        trim_axes: bool = True,
        quantile: float | None = None,
        fig: plt.Figure | None = None,
        ax: plt.Axes | None = None,
        legend_ax: plt.Axes | None = None,
        null_linestyle: str = "dotted",
        legend_n_cols: int | None = None,
    ) -> tuple[plt.Figure, plt.Axes, plt.Figure | None, plt.Axes | None]:
        """
        Temporal contributions of different source types for positive/negative contrasts.

        Args:
            contrast: The contrast of the stimulus.
            angle: The angle of the stimulus.
            t_start: Start time for the plot.
            t_end: End time for the plot.
            fontsize: Font size for labels and titles.
            linewidth: Line width for traces.
            legend: Whether to show the legend.
            legend_standalone: Whether to create a standalone legend.
            legend_figsize_cm: Figure size for the standalone legend.
            legend_n_rows: Number of rows for the standalone legend.
            max_figure_height_cm: Maximum figure height in centimeters.
            panel_height_cm: Height of each panel in centimeters.
            max_figure_width_cm: Maximum figure width in centimeters.
            panel_width_cm: Width of each panel in centimeters.
            model_average: Whether to plot the model average.
            highlight_mean: Whether to highlight the mean trace.
            sum_exc_inh: Whether to sum excitatory and inhibitory contributions.
            only_sum: Whether to only plot the summed contributions.
            hide_source_types: Source types to hide or "auto".
            hide_source_types_bins: Number of bins for auto-hiding.
            hide_source_types_cut_off_edge: Cut-off edge for auto-hiding.
            hide_source_types_mode: Mode for auto-hiding source types.
            hide_yaxis: Whether to hide the y-axis.
            trim_axes: Whether to trim the axes.
            quantile: Quantile for shading.
            fig: Existing figure to use.
            ax: Existing axes to use.
            legend_ax: Existing axes for the standalone legend.
            null_linestyle: Linestyle for null direction traces.
            legend_n_cols: Number of columns for the standalone legend.

        Returns:
            Figure, axes, and legend axes objects.

        Example:
            ```
            view = MovingEdgeCurrentView(...)
            fig, ax = view.plot_temporal_contributions_pc_nc(
                contrast=1.0,
                angle=0,
                t_start=0,
                t_end=1,
                fontsize=5,
                linewidth=0.25,
                legend=True
            )
            ```
        """
        if fig is None and ax is None:
            figsize = figsize_from_n_items(
                1,
                max_figure_height_cm=max_figure_height_cm,
                panel_height_cm=panel_height_cm,
                max_figure_width_cm=max_figure_width_cm,
                panel_width_cm=panel_width_cm,
            )
            fig, axes = figsize.axis_grid(hspace=0.0, wspace=0, fontsize=fontsize)
            ax = axes[0]
        cv_pd = (
            self.at_contrast(contrast)
            .at_angle(angle)
            .between_seconds(t_start, t_end)
            .sum_over_cells()
        )
        cv_nd = (
            self.at_contrast(contrast)
            .at_angle((angle - 180) % 360)
            .between_seconds(t_start, t_end)
            .sum_over_cells()
        )

        source_types = (
            self.at_contrast(contrast)
            .at_angle([angle, (angle - 180) % 360])
            .between_seconds(t_start, t_end)
            .filter_source_types(
                hide_source_types,
                hide_source_types_bins,
                hide_source_types_cut_off_edge,
                hide_source_types_mode,
            )
        )

        color_source_types = (
            self.at_contrast(contrast)
            .at_angle([angle, (angle - 180) % 360])
            .between_seconds(t_start, t_end)
            .filter_source_types(
                None,
                hide_source_types_bins,
                hide_source_types_cut_off_edge,
                hide_source_types_mode,
            )
        )
        cv_pd.init_colors(color_source_types)
        cv_nd.init_colors(color_source_types)

        def plot_mean_trace(
            time, trace, label, color, zorder, linestyle="solid", ax=None, fig=None
        ):
            ax.plot(
                time,
                trace,
                label=label,
                color=color,
                zorder=zorder,
                linestyle=linestyle,
            )

        def plot_individual_traces(
            traces, time, color, zorder, label, linestyle="solid", legend=None
        ):
            if not only_sum and not model_average:
                plots.traces(
                    traces,
                    time,
                    mean_color=color,
                    color=color,
                    linewidth=linewidth,
                    zorder_traces=0,
                    zorder_mean=zorder,
                    fontsize=fontsize,
                    null_line=True,
                    highlight_mean=highlight_mean,
                    fig=fig,
                    ax=ax,
                    legend=legend or label,
                    linestyle=linestyle,
                )

        def plot_quantile(traces, time, color, zorder, linestyle="solid"):
            if quantile:
                Q = np.quantile(traces, quantile, axis=0)
                ax.fill_between(
                    time,
                    Q[0],
                    Q[1],
                    facecolor=adapt_color_alpha(color, 0.1),
                    edgecolor=color,
                    linewidth=0.25,
                    linestyle=linestyle,
                    zorder=zorder - 1,
                )

        def plot_summed_trace(time, trace, label, color, zorder, linestyle="solid"):
            if np.any(trace):
                ax.plot(
                    time,
                    trace,
                    label=label,
                    color=color,
                    zorder=zorder,
                    linestyle=linestyle,
                )

        def get_summed_traces(signs, source_types, cv_pd, cv_nd):
            # sum over cell types then average over models
            exc_pd = np.zeros(cv_pd.shape)
            inh_pd = np.zeros(cv_pd.shape)
            exc_nd = np.zeros(cv_nd.shape)
            inh_nd = np.zeros(cv_nd.shape)
            # sum over cell types
            for source_type in source_types:
                if signs[source_type] == 1:
                    exc_pd += cv_pd[source_type][:]  # (1, n_models, 1, n_timesteps)
                    exc_nd += cv_nd[source_type][:]
                else:
                    inh_pd += cv_pd[source_type][:]
                    inh_nd += cv_nd[source_type][:]
            # (n_models, n_timesteps)
            return (
                exc_pd.squeeze(),
                inh_pd.squeeze(),
                exc_nd.squeeze(),
                inh_nd.squeeze(),
            )

        for source_type in source_types:
            if model_average and not only_sum:
                # mean traces solid for PD and dashed for ND
                color = cv_pd.color(source_type)

                plot_mean_trace(
                    cv_pd.time,
                    cv_pd[source_type][:].squeeze(axis=-2).T.mean(axis=1),
                    source_type,
                    color,
                    cv_pd.zorder(source_types, source_type),
                    ax=ax,
                    fig=fig,
                )
                plot_mean_trace(
                    cv_nd.time,
                    cv_nd[source_type][:].squeeze(axis=-2).T.mean(axis=1),
                    source_type,
                    color,
                    linestyle=null_linestyle,
                    zorder=cv_pd.zorder(source_types, source_type),
                    ax=ax,
                    fig=fig,
                )

            elif not model_average and not only_sum:
                # individual traces
                plot_individual_traces(
                    cv_pd[source_type][:].squeeze(axis=-2),
                    cv_pd.time,
                    cv_pd.color(source_type),
                    cv_pd.zorder(source_types, source_type),
                    source_type,
                )
                plot_individual_traces(
                    cv_nd[source_type][:].squeeze(axis=-2),
                    cv_nd.time,
                    cv_pd.color(source_type),
                    cv_pd.zorder(source_types, source_type),
                    source_type,
                    linestyle=null_linestyle,
                    legend="null direction",
                )

            # quantiles
            plot_quantile(
                cv_pd[source_type][:].squeeze(axis=-2),
                cv_pd.time,
                cv_pd.color(source_type),
                cv_pd.zorder(source_types, source_type),
                linestyle="solid",
            )
            plot_quantile(
                cv_nd[source_type][:].squeeze(axis=-2),
                cv_nd.time,
                cv_pd.color(source_type),
                cv_pd.zorder(source_types, source_type),
                linestyle=null_linestyle,
            )
        if sum_exc_inh or only_sum:
            # plot summed traces
            signs = cv_pd.signs()
            exc_pd, inh_pd, exc_nd, inh_nd = get_summed_traces(
                signs, source_types, cv_pd, cv_nd
            )
            plot_summed_trace(
                cv_pd.time,
                exc_pd.mean(axis=0),
                "excitatory",
                (0.931, 0.0, 0.0, 1.0),
                zorder=2000,
            )
            plot_quantile(
                exc_pd,
                cv_pd.time,
                (0.931, 0.0, 0.0, 1.0),
                zorder=0,
                linestyle="solid",
            )
            plot_summed_trace(
                cv_nd.time,
                exc_nd.mean(axis=0),
                "excitatory",
                (0.931, 0.0, 0.0, 1.0),
                zorder=2000,
                linestyle=null_linestyle,
            )
            plot_quantile(
                exc_nd,
                cv_pd.time,
                (0.931, 0.0, 0.0, 1.0),
                zorder=0,
                linestyle=null_linestyle,
            )
            plot_summed_trace(
                cv_pd.time,
                inh_pd.mean(axis=0),
                "inhibitory",
                (0.0, 0.0, 0.849, 1.0),
                zorder=2000,
            )
            plot_quantile(
                inh_pd,
                cv_pd.time,
                (0.0, 0.0, 0.849, 1.0),
                zorder=0,
                linestyle="solid",
            )
            plot_summed_trace(
                cv_nd.time,
                inh_nd.mean(axis=0),
                "inhibitory",
                (0.0, 0.0, 0.849, 1.0),
                zorder=2000,
                linestyle=null_linestyle,
            )
            plot_quantile(
                inh_nd,
                cv_pd.time,
                (0.0, 0.0, 0.849, 1.0),
                zorder=0,
                linestyle=null_linestyle,
            )

        if legend:
            ax.legend(
                fontsize=fontsize,
                ncols=1,
                bbox_to_anchor=(1.05, 1),
                loc="upper left",
                borderaxespad=0.0,
            )
        else:
            ax.legend().set_visible(False)

        ax.set_xlabel("time (s)", fontsize=fontsize)
        #         ax.set_ylabel("current (a.u.)", fontsize=fontsize)

        if hide_yaxis:
            plt_utils.rm_spines(ax, ("left",))

        if trim_axes:
            plt_utils.trim_axis(ax)

        if legend_standalone:
            handles, labels = ax.get_legend_handles_labels()
            nd_handle = Line2D(
                [0], [0], color="k", lw=1, label="null direction", ls=null_linestyle
            )
            legend_n_rows = legend_n_rows or len(labels) + 1
            # legend_n_cols = (len(labels) + 1) // legend_n_rows
            legend_fig, legend_ax = plt_utils.standalone_legend(
                [*labels[::2], "null direction"],
                None,
                [*handles[::2], nd_handle],
                fontsize=fontsize,
                n_cols=legend_n_cols,
                handlelength=2,
                columnspacing=0.8,
                labelspacing=0.25,
                figsize=cm_to_inch(legend_figsize_cm),
                fig=fig if legend_ax is not None else None,
                ax=legend_ax,
            )
            return fig, ax, legend_fig, legend_ax
        return fig, ax, None, None

    def get_temporal_contributions(
        self,
        contrast: float,
        angle: float,
        t_start: float = 0,
        t_end: float = 1,
        hide_source_types: str | list | None = "auto",
        hide_source_types_bins: int = 5,
        hide_source_types_cut_off_edge: int = 1,
        hide_source_types_mode: str = "below_cut_off",
        summed_traces: bool = False,
    ) -> tuple[
        "MovingEdgeCurrentView" | np.ndarray,
        "MovingEdgeCurrentView" | np.ndarray,
        list[str],
        list[str],
    ]:
        cv_pd = (
            self.at_contrast(contrast)
            .at_angle(angle)
            .between_seconds(t_start, t_end)
            .sum_over_cells()
        )
        cv_nd = (
            self.at_contrast(contrast)
            .at_angle((angle - 180) % 360)
            .between_seconds(t_start, t_end)
            .sum_over_cells()
        )

        source_types = (
            self.at_contrast(contrast)
            .at_angle([angle, (angle - 180) % 360])
            .between_seconds(t_start, t_end)
            .filter_source_types(
                hide_source_types,
                hide_source_types_bins,
                hide_source_types_cut_off_edge,
                hide_source_types_mode,
            )
        )

        color_source_types = (
            self.at_contrast(contrast)
            .at_angle([angle, (angle - 180) % 360])
            .between_seconds(t_start, t_end)
            .filter_source_types(
                None,
                hide_source_types_bins,
                hide_source_types_cut_off_edge,
                hide_source_types_mode,
            )
        )
        cv_pd.init_colors(color_source_types)
        cv_nd.init_colors(color_source_types)

        def get_summed_traces(signs, source_types, cv_pd, cv_nd):
            # sum over cell types then average over models
            exc_pd = np.zeros(cv_pd.shape)
            inh_pd = np.zeros(cv_pd.shape)
            exc_nd = np.zeros(cv_nd.shape)
            inh_nd = np.zeros(cv_nd.shape)
            # sum over cell types
            for source_type in source_types:
                if signs[source_type] == 1:
                    exc_pd += cv_pd[source_type][:]  # (1, n_models, 1, n_timesteps)
                    exc_nd += cv_nd[source_type][:]
                else:
                    inh_pd += cv_pd[source_type][:]
                    inh_nd += cv_nd[source_type][:]
            # (n_models, n_timesteps)
            return (
                exc_pd.squeeze(),
                inh_pd.squeeze(),
                exc_nd.squeeze(),
                inh_nd.squeeze(),
            )

        if summed_traces:
            exc_pd, inh_pd, exc_nd, inh_nd = get_summed_traces(
                cv_pd.signs(), source_types, cv_pd, cv_nd
            )
            # return exc_pd, inh_pd, exc_nd, inh_nd, source_types, color_source_types
            return exc_pd, inh_pd, exc_nd, inh_nd

        return cv_pd, cv_nd, source_types, color_source_types

    def get_response(
        self,
        contrast: float,
        angle: float,
        t_start: float = 0,
        t_end: float = 1,
        model_average: bool = True,
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        r_pd = (
            self.at_angle(angle)
            .at_contrast(contrast)
            .between_seconds(t_start, t_end)
            .responses.squeeze(axis=-2)
        )
        r_nd = (
            self.at_angle((angle - 180) % 360)
            .at_contrast(contrast)
            .between_seconds(t_start, t_end)
            .responses.squeeze(axis=-2)
        )

        if model_average:
            return (
                r_pd.mean(axis=0),
                r_nd.mean(axis=0),
                self.between_seconds(t_start, t_end).time,
            )

        return r_pd, r_nd, self.between_seconds(t_start, t_end).time


def reset_index(rfs: ReceptiveFields, inplace: bool = False) -> ReceptiveFields:
    if inplace:
        for source_type in rfs.source_types:
            edges = rfs[source_type]
            if isinstance(edges, pd.DataFrame):
                rfs[source_type] = edges.reset_index(drop=True)
        return rfs
    else:
        new = rfs.deepcopy()
        return reset_index(new, inplace=True)


def at_position(
    rfs: ReceptiveFields,
    u: float | None = None,
    v: float | None = None,
    central: bool = True,
    inplace: bool = False,
) -> ReceptiveFields:
    if inplace:
        for source_type in rfs.source_types:
            edges = rfs[source_type]
            if isinstance(edges, pd.DataFrame):
                if central:
                    rfs[source_type] = edges.iloc[
                        [np.argmin(np.abs(edges.source_u) + np.abs(edges.source_v))]
                    ]
                else:
                    rfs[source_type] = edges[
                        (edges.source_u == u) & (edges.source_v == v)
                    ]
        return rfs
    else:
        new = rfs.deepcopy()
        return at_position(new, u, v, inplace=True)



"""Cartesian frame animation."""

from time import sleep
from typing import List, Optional

import matplotlib.pyplot as plt
import numpy as np

from ..visualization import plt_utils
from .animations import Animation


class Imshow(Animation):
    """Animates an array of images using imshow.

    Args:
        images: Array of images to animate (n_samples, n_frames, height, width).
        fig: Existing Figure instance or None.
        ax: Existing Axis instance or None.
        update: Whether to update the canvas after an animation step.
            Must be False if this animation is composed with others.
        figsize: Size of the figure.
        sleep: Time to sleep between frames.
        **kwargs: Additional arguments passed to plt.imshow.

    Attributes:
        fig (plt.Figure): The figure object.
        ax (plt.Axes): The axes object.
        kwargs (dict): Additional arguments for imshow.
        update (bool): Whether to update the canvas after each step.
        n_samples (int): Number of samples in the images array.
        frames (int): Number of frames in each sample.
        images (np.ndarray): Array of images to animate.
        sleep (float): Time to sleep between frames.
        img (plt.AxesImage): The image object created by imshow.

    Note:
        The `images` array should have shape (n_samples, n_frames, height, width).
    """

    def __init__(
        self,
        images: np.ndarray,
        fig: Optional[plt.Figure] = None,
        ax: Optional[plt.Axes] = None,
        update: bool = True,
        figsize: List[int] = [1, 1],
        sleep: float = 0.01,
        **kwargs,
    ) -> None:
        super().__init__()
        self.fig, self.ax = plt_utils.init_plot(
            figsize=figsize,
            fig=fig,
            ax=ax,
            position=[0, 0, 1, 1],
            set_axis_off=True,
        )
        self.kwargs = kwargs
        self.update = update
        self.n_samples, self.frames = images.shape[:2]
        self.images = images
        self.sleep = sleep
        super().__init__(None, self.fig)

    def init(self, frame: int = 0) -> None:
        """Initialize the animation.

        Args:
            frame: The initial frame to display.
        """
        self.img = self.ax.imshow(self.images[self.batch_sample, frame], **self.kwargs)
        if self.sleep is not None:
            sleep(self.sleep)

    def animate(self, frame: int) -> None:
        """Animate a single frame.

        Args:
            frame: The frame number to animate.
        """
        self.img.set_data(self.images[self.batch_sample, frame])

        if self.update:
            self.update_figure()

        if self.sleep is not None:
            sleep(self.sleep)



from types import SimpleNamespace

import numpy as np

from flyvis.network import ensemble_view as ensemble_view_module
from flyvis.network.ensemble_view import EnsembleView


class _FakeFRI:
    def __init__(self, values, cell_types, filtered_cell_types=None):
        self.values = values
        self.cell_type = SimpleNamespace(values=np.array(cell_types))
        self.filtered_cell_types = filtered_cell_types
        self.custom = SimpleNamespace(where=self._where)

    def _where(self, cell_type):
        if self.filtered_cell_types is not None:
            return _FakeFRI(self.values, self.filtered_cell_types)
        return _FakeFRI(self.values, list(cell_type))


class _DummyView:
    def flash_responses(self):
        return "responses"

    def task_error(self):
        return SimpleNamespace(values=np.array([0.5, 0.1, 0.2]))


def test_flash_response_index_aligns_labels_with_filtered_data(monkeypatch):
    requested = ["Mi1", "Tm3", "CT1(M10)"]
    filtered = ["CT1(M10)", "Mi1", "Tm3"]
    fake_fris = _FakeFRI(np.ones((3, 2, 1)), requested, filtered_cell_types=filtered)

    monkeypatch.setattr(
        ensemble_view_module, "flash_response_index", lambda *_args, **_kwargs: fake_fris
    )

    captured = SimpleNamespace(fris=None, cell_types=None, sorted_type_list=None)

    def _fake_plot_fris(fris, cell_types, **kwargs):
        captured.fris = fris
        captured.cell_types = list(cell_types)
        captured.sorted_type_list = kwargs.get("sorted_type_list")
        return "fig", "ax"

    monkeypatch.setattr(ensemble_view_module, "plot_fris", _fake_plot_fris)

    fig, ax = EnsembleView.flash_response_index(_DummyView(), cell_types=requested)

    assert (fig, ax) == ("fig", "ax")
    assert captured.fris.shape == (3, 2, 1)
    assert captured.cell_types == filtered
    assert captured.sorted_type_list == requested



"""Solvers for training, testing, checkpointing and recovering of networks."""

import logging
import time
from dataclasses import dataclass
from typing import Dict, Optional, Protocol, Union

import numpy as np
import torch
from datamate import Directory, Namespace
from toolz import valfilter, valmap
from torch import nn

from flyvis.network.directories import NetworkDir
from flyvis.network.network import Network
from flyvis.task.tasks import Task
from flyvis.utils.chkpt_utils import (
    recover_decoder,
    recover_network,
    recover_optimizer,
    recover_penalty_optimizers,
    resolve_checkpoints,
)
from flyvis.utils.tensor_utils import asymmetric_weighting

logging = logging.getLogger(__name__)

__all__ = ["MultiTaskSolver", "Penalty", "HyperParamScheduler"]


class SolverProtocol(Protocol):
    """SolverProtocol implements training, testing, checkpointing etc. of networks."""

    name: str
    config: Optional[Union[dict, Namespace]]

    def __init__(
        self, name: str = "", config: Optional[Union[dict, Namespace]] = None
    ) -> None: ...

    dir: Directory = None
    network: Network = None
    decoder: Dict[str, nn.Module] = None
    task: Task = None
    optimizer: object = None
    penalty: object = None
    scheduler: object = None

    def train(self) -> None: ...

    def checkpoint(self) -> None: ...

    def test(self) -> None: ...

    def recover(self) -> None: ...


class MultiTaskSolver:
    """Implements training, testing, checkpoint, recovering of flyvis networks.

    Gives access to the network, decoder, task, optimizer, penalty and scheduler and
    the directory where the results are stored.

    Args:
        name: Name of the solver.
        config: Configuration for the solver.
        init_network: Whether to initialize the network. Defaults to True.
        init_decoder: Whether to initialize the decoder. Defaults to True.
        init_task: Whether to initialize the task. Defaults to True.
        init_optim: Whether to initialize the optimizer. Defaults to True.
        init_penalties: Whether to initialize penalties. Defaults to True.
        init_scheduler: Whether to initialize the scheduler. Defaults to True.
        delete_if_exists: Whether to delete existing directory. Defaults to False.

    Attributes:
        dir (NetworkDir): Directory where results are stored.
        network (Network): The neural network.
        decoder (Dict[str, nn.Module]): The decoder modules.
        task (Task): The task being solved.
        optimizer (torch.optim.Optimizer): The optimizer.
        penalty (Penalty): The penalty object.
        scheduler (HyperParamScheduler): The hyperparameter scheduler.

    Example:
        ```python
        from flyvis.utils.config_utils import get_default_config
        # Note: the config is typically defined through the command line.
        config = get_default_config(overrides=["task_name=flow",
                                               "ensemble_and_network_id=0"])
        solver = MultiTaskSolver("test", config)
        solver.train()
        ```
    """

    def __init__(
        self,
        name: str = "",
        config: Optional[Union[dict, Namespace]] = None,
        init_network: bool = True,
        init_decoder: bool = True,
        init_task: bool = True,
        init_optim: bool = True,
        init_penalties: bool = True,
        init_scheduler: bool = True,
        delete_if_exists: bool = False,
    ) -> None:
        name = name or config["network_name"]
        assert isinstance(name, str), "Provided name argument is not a string."
        self.dir = NetworkDir(
            name, {**(config or {}), **dict(delete_if_exists=delete_if_exists)}
        )

        self.path = self.dir.path

        self.config = self.dir.config

        self.iteration = 0
        self._val_loss = float("inf")
        self.checkpoint_path = self.dir.path / "chkpts"
        checkpoints = resolve_checkpoints(self.dir)
        self.checkpoints = checkpoints.indices
        self._last_chkpt_ind = -1
        self._curr_chkpt_ind = -1

        self._initialized = self._init_solver(
            init_network=init_network,
            init_decoder=init_decoder,
            init_task=init_task,
            init_optim=init_optim,
            init_penalties=init_penalties,
            init_scheduler=init_scheduler,
        )

        logging.info("Initialized solver.")
        logging.info(repr(self.config))

    def _init_solver(
        self,
        init_network: bool = False,
        init_decoder: bool = False,
        init_task: bool = False,
        init_optim: bool = False,
        init_penalties: bool = False,
        init_scheduler: bool = False,
    ) -> list:
        """Initialize solver components.

        Args:
            init_network: Whether to initialize the network.
            init_decoder: Whether to initialize the decoder.
            init_task: Whether to initialize the task.
            init_optim: Whether to initialize the optimizer.
            init_penalties: Whether to initialize penalties.
            init_scheduler: Whether to initialize the scheduler.

        Returns:
            A list of initialized components.
        """
        initialized = []

        if init_network:
            self.network = Network(**self.config.network)
            initialized.append("network")

        if init_task:
            self.task = Task(**self.config.task)
            initialized.append("task")

            if init_decoder:
                self.decoder = self.task.init_decoder(self.network.connectome)
                initialized.append("decoder")

        if init_optim:
            self.optimizer = self._init_optimizer(
                self.config.optim, self.network, self.decoder
            )
            initialized.append("optim")

        if init_penalties:
            self.penalty = Penalty(self.config.penalizer, self.network)
            initialized.append("penalties")

        if init_scheduler:
            self.scheduler = HyperParamScheduler(
                self.config.scheduler,
                self.network,
                self.task,
                self.optimizer,
                self.penalty,
            )
            self.scheduler(self.iteration)
            initialized.append("scheduler")

        return initialized

    @staticmethod
    def _init_optimizer(
        optim: Namespace, network: Network, decoder: Optional[Dict[str, nn.Module]]
    ) -> torch.optim.Optimizer:
        """Initializes the optimizer for network and decoder.

        Args:
            optim: Optimizer configuration.
            network: The neural network.
            decoder: The decoder modules.

        Returns:
            The initialized optimizer.
        """

        def decoder_parameters(decoder: Dict[str, nn.Module]):
            """Returns decoder parameters."""
            params = []
            for nn_module in decoder.values():
                params.append(
                    dict(
                        params=[w for w in nn_module.parameters()],
                        **config.optim_dec,
                    )
                )
            return params

        config = optim.deepcopy()

        optim_type = config.pop("type", "Adam")
        optim = torch.optim.__dict__[optim_type]
        logging.info("Initializing %s for network and decoder.", optim.__name__)

        param_groups = [dict(params=network.parameters(), **config.optim_net)]

        if decoder:
            param_groups.extend(decoder_parameters(decoder))

        return optim(param_groups)

    def train(self, overfit: bool = False, initial_checkpoint: bool = True) -> None:
        """Trains the network by backprop through time.

        Args:
            overfit: If true, the dataloader is substituted by a
                single-sequence loader and augmentation is turned off.
            initial_checkpoint: Whether to create an initial checkpoint when debugging.

        Raises:
            OverflowError: If the activity or loss reports NaN values for more
                than 100 iterations.

        Stores:
            ```bash
            dir / loss.h5
            dir / loss_<task>.h5
            dir / activity.h5
            dir / activity_min.h5
            dir / activity_max.h5
            ```
        """
        # return if iterations have already been trained.
        if self.iteration >= self.task.n_iters:
            return

        # to debug code within the training loop the initial checkpoint should be
        # disabled
        if initial_checkpoint:
            self.checkpoint()

        logging.info("Starting training.")
        # The overfit_data dataloader only contains a single sequence and
        # this is to debug the model architecture, configs etc.
        dataloader = self.task.overfit_data if overfit else self.task.train_data
        # For overfitting we also turn the augmentation off.
        augment = not overfit

        # The number of full presentations of the training data is derived from the
        # preset number of training iterations, the length of the dataloader and the
        # current iteration.
        n_epochs = np.ceil((self.task.n_iters - self.iteration) / len(dataloader)).astype(
            int
        )

        # This is after how many epochs the training states are checkpointed.
        chkpt_every_epoch = self.config.scheduler.chkpt_every_epoch

        logging.info("Training for %s epochs.", n_epochs)
        logging.info("Checkpointing every %s epochs.", chkpt_every_epoch)

        # Initialize data structures to store the loss and activity over iterations.
        loss_over_iters = []
        activity_over_iters = []
        activity_min_over_iters = []
        activity_max_over_iters = []
        loss_per_task = {f"loss_{task}": [] for task in self.task.dataset.tasks}

        start_time = time.time()
        with self.task.dataset.augmentation(augment):
            for epoch in range(n_epochs):
                # The default is to compute a steady state for each epoch, then
                # it's computed here. Note: unless done per iteration, parameter updates
                # within epochs are not considered in the steady state.
                steady_state = self.network.steady_state(
                    t_pre=self.config.get("t_pre_train", 0.5),
                    dt=self.task.dataset.dt,
                    batch_size=dataloader.batch_size,
                    value=0.5,
                )

                for _, data in enumerate(dataloader):

                    def handle_batch(data, steady_state):
                        """Closure to free memory by garbage collector effectively."""

                        # Resets the stimulus buffer (samples, frames, neurons).
                        n_samples, n_frames, _, _ = data["lum"].shape
                        self.network.stimulus.zero(n_samples, n_frames)

                        # Add batch of hex-videos (#frames, #samples, #hexals) as
                        # photorecptor stimuli.
                        self.network.stimulus.add_input(data["lum"])

                        # Reset gradients.
                        self.optimizer.zero_grad()

                        # Run stimulus through network.
                        activity = self.network(
                            self.network.stimulus(),
                            self.task.dataset.dt,
                            state=steady_state,
                        )

                        losses = {task: 0 for task in self.task.dataset.tasks}
                        for task in self.task.dataset.tasks:
                            y = data[task]
                            y_est = self.decoder[task](activity)

                            # to pass additional kwargs to the loss function, from
                            # the data batch from the dataset
                            losses[task] = self.task.loss(
                                y_est, y, task, **data.get("loss_kwargs", {})
                            )

                        # Sum all task losses. The weighting of the tasks is done in the
                        # loss function.
                        loss = sum(losses.values())

                        # Compute gradients.
                        loss.backward(retain_graph=True)
                        # Update parameters.
                        self.optimizer.step()

                        # Activity and parameter dependent penalties.
                        self.penalty(activity=activity, iteration=self.iteration)

                        # Log results.
                        loss = loss.detach().cpu()
                        for task in self.task.dataset.tasks:
                            loss_per_task[f"loss_{task}"].append(
                                losses[task].detach().cpu().item()
                            )
                        loss_over_iters.append(loss.item())
                        activity = activity.detach().cpu()
                        mean_activity = activity.mean()
                        activity_over_iters.append(mean_activity.item())
                        activity_min_over_iters.append(activity.min().item())
                        activity_max_over_iters.append(activity.max().item())
                        return loss, mean_activity

                    # Call closure.
                    loss, mean_activity = handle_batch(data, steady_state)

                    # Increment iteration count.
                    self.iteration += 1

                # Interrupt training if the network explodes.
                if torch.isnan(loss) or torch.isnan(mean_activity):
                    logging.warning("Network exploded.")
                    raise OverflowError("Invalid values encountered in trace.")

                # The scheduling of hyperparams are functions of the iteration
                # however, we allow steps only after full presentations of the data.
                if epoch + 1 != n_epochs:
                    self.scheduler(self.iteration)
                    logging.info("Scheduled paremeters for iteration %s.", self.iteration)

                # Checkpointing.
                if (epoch % chkpt_every_epoch == 0) or (epoch + 1 == n_epochs):
                    self.dir.loss = loss_over_iters
                    self.dir.activity = activity_over_iters
                    self.dir.activity_min = activity_min_over_iters
                    self.dir.activity_max = activity_max_over_iters

                    for task in self.task.dataset.tasks:
                        self.dir[f"loss_{task}"] = loss_per_task[f"loss_{task}"]

                    self.checkpoint()

                logging.info("Finished epoch.")

        time_elapsed = time.time() - start_time
        time_trained = self.dir.time_trained[()] if "time_trained" in self.dir else 0
        self.dir.time_trained = time_elapsed + time_trained
        logging.info("Finished training.")

    def checkpoint(self) -> None:
        """Creates a checkpoint.

        Validates on the validation data calling ~self.test.
        Validates on a training batch calling ~self.track_batch.
        Stores a checkpoint of the network, decoder and optimizer parameters using
        pytorch's pickle function.

        Stores:
            ```bash
            dir / chkpt_index.h5  # (List): numerical identifier of the checkpoint.
            dir / chkpt_iter.h5  # (List): iteration at which this checkpoint was
                                 # recorded.
            dir / best_chkpt_index.h5  # (int): chkpt index at which the val loss is
                                       # minimal.
            dir / dt.h5  # (float): the current time constant of the dataset.
            dir / chkpts / chkpt_<chkpt_index>  # (dict): the state dicts of the network,
                                                # decoder and optimizer.
            ```
        """
        self._last_chkpt_ind += 1
        self._curr_chkpt_ind += 1

        # Tracking of validation loss and training batch loss.
        logging.info("Test on validation data.")
        val_loss = self.test(
            dataloader=self.task.val_data, subdir="validation", track_loss=True
        )
        logging.info("Test on validation batch.")
        _ = self.test(
            dataloader=self.task.val_batch, subdir="validation_batch", track_loss=True
        )
        logging.info("Test on training data.")
        _ = self.test(dataloader=self.task.train_data, subdir="training", track_loss=True)
        logging.info("Test on training batch.")
        _ = self.test(
            dataloader=self.task.train_batch, subdir="training_batch", track_loss=True
        )

        logging.info("Saving state dicts.")
        # Store state of pytorch modules.
        nn_state_dict = self.network.state_dict()
        dec_state_dict = {}
        if self.decoder:
            dec_state_dict = valmap(lambda x: x.state_dict(), self.decoder)
        chkpt = {
            "network": nn_state_dict,
            "decoder": dec_state_dict,
            "optim": self.optimizer.state_dict(),
            "time": time.ctime(),
            "val_loss": val_loss,
            "iteration": self.iteration - 1,
            "dt": self.task.dataset.dt,
        }
        if hasattr(self, "penalty"):
            chkpt.update(self.penalty._chkpt())
        torch.save(chkpt, self.checkpoint_path / f"chkpt_{self._last_chkpt_ind:05}")

        # Append chkpt index.
        self.checkpoints.append(self._last_chkpt_ind)
        self.dir.extend("chkpt_index", [self._last_chkpt_ind])
        self.dir.extend("chkpt_iter", [self.iteration - 1])
        self.dir.dt = self.task.dataset.dt

        # Overwrite best val loss.
        if val_loss < self._val_loss:
            self.dir.best_chkpt_index = self._last_chkpt_ind
            self._val_loss = val_loss

        logging.info("Checkpointed.")

    @torch.no_grad()
    def test(
        self,
        dataloader: torch.utils.data.DataLoader,
        subdir: str = "validation",
        track_loss: bool = False,
        t_pre: float = 0.25,
    ) -> float:
        """Tests the network on a given dataloader.

        Args:
            dataloader: Data to test on.
            subdir: Name of subdirectory. Defaults to 'validation'.
            track_loss: Whether to store the loss in dir.subdir.
            t_pre: Warmup time before the stimulus starts.

        Returns:
            Validation loss.

        Stores:
            ```bash
            dir.<subdir>.loss_<task>  # (List): Loss per task, averaged over whole
                                      # dataset.
            dir.<subdir>.iteration  # (List): Iteration when this was called.
            dir.<subdir>.loss  # (List): Average loss over tasks.
            ```
        """
        self._eval()
        logging.info("Test")

        # Update hypterparams.
        self.scheduler(self.iteration)

        initial_state = self.network.steady_state(
            t_pre=t_pre,
            dt=self.task.dataset.dt,
            batch_size=dataloader.batch_size,
            value=0.5,
        )
        losses = {task: () for task in self.task.dataset.tasks}  # type: Dict[str, Tuple]

        with self.task.dataset.augmentation(False):
            for _, data in enumerate(dataloader):
                n_samples, n_frames, _, _ = data["lum"].shape
                self.network.stimulus.zero(n_samples, n_frames)

                self.network.stimulus.add_input(data["lum"])

                activity = self.network(
                    self.network.stimulus(),
                    self.task.dataset.dt,
                    state=initial_state,
                )

                for task in self.task.dataset.tasks:
                    y = data[task]
                    y_est = self.decoder[task](activity)

                    losses[task] += (
                        self.task.loss(y_est, y, task, **data.get("loss_kwargs", {}))
                        .detach()
                        .cpu()
                        .item(),
                    )

        # track loss per task.
        avg_loss_per_task = {}
        for task in self.task.dataset.tasks:
            # average the loss over the whole dataset
            avg_loss_per_task[task] = np.mean(losses[task])
            if track_loss:
                self.dir[subdir].extend("loss" + "_" + task, [avg_loss_per_task[task]])

        # average the loss over all tasks with equal weight
        summed_loss = sum(avg_loss_per_task.values())
        val_loss = summed_loss / len(avg_loss_per_task)

        if track_loss:
            self.dir[subdir].extend("iteration", [self.iteration])
            self.dir[subdir].extend("loss", [val_loss])

        self._train()

        return val_loss

    def _train(self) -> None:
        """Sets nn.Modules to train state."""
        self.network.train()
        if self.decoder is not None:
            for decoder in self.decoder.values():
                decoder.train()

    def _eval(self) -> None:
        """Sets nn.Modules to eval state."""
        self.network.eval()
        if self.decoder is not None:
            for decoder in self.decoder.values():
                decoder.eval()

    def recover(
        self,
        network: bool = True,
        decoder: bool = True,
        optimizer: bool = True,
        penalty: bool = True,
        checkpoint: Union[int, str] = "best",
        validation_subdir: str = "validation",
        loss_file_name: str = "loss",
        strict: bool = True,
        force: bool = False,
    ) -> None:
        """Recovers the solver state from a checkpoint.

        Args:
            network: Recover network parameters.
            decoder: Recover decoder parameters.
            optimizer: Recover optimizer parameters.
            penalty: Recover penalty parameters.
            checkpoint: Index of the checkpoint to recover.
            validation_subdir: Name of the subdir to base the best checkpoint on.
            loss_file_name: Name of the loss to base the best checkpoint on.
            strict: Whether to load the state dict of the decoders strictly.
            force: Force recovery of checkpoint if _curr_chkpt_ind is already
                the same as the checkpoint index.
        """
        checkpoints = resolve_checkpoints(
            self.dir, checkpoint, validation_subdir, loss_file_name
        )

        if checkpoint.index is None or not any((network, decoder, optimizer, penalty)):
            logging.info("No checkpoint found. Continuing with initialized parameters.")
            return

        if checkpoints.index == self._curr_chkpt_ind and not force:
            logging.info("Checkpoint already recovered.")
            return

        # Set the current and last checkpoint index. New checkpoints incrementally
        # increase the last checkpoint index.
        self._last_chkpt_ind = checkpoints.indices[-1]
        self._curr_chkpt_ind = checkpoints.index

        # Load checkpoint data.
        state_dict = torch.load(checkpoints.path)
        logging.info(f"Checkpoint {checkpoints.path} loaded.")

        self.iteration = state_dict.get("iteration", None)

        if "scheduler" in self._initialized:
            # Set the scheduler to the right iteration.
            self.scheduler(self.iteration)

        # The _val_loss variable is used to keep track of the best checkpoint according
        # to the evaluation routine during training.
        self._val_loss = state_dict.pop("val_loss", float("inf"))

        if network and "network" in self._initialized:
            recover_network(self.network, state_dict)
        if decoder and "decoder" in self._initialized:
            recover_decoder(self.decoder, state_dict, strict=strict)
        if optimizer and "optim" in self._initialized:
            recover_optimizer(self.optimizer, state_dict)
        if penalty and "penalties" in self._initialized:
            recover_penalty_optimizers(self.penalty.optimizers, state_dict)

        logging.info("Recovered modules.")


class Penalty:
    """Penalties on specific parameters.

    Args:
        penalty: Penalty configuration.
        network: The neural network.

    Note: Default config in config/penalizer/penalizer.yaml
        ```yaml
        activity_penalty:
            activity_baseline: 5.0
            activity_penalty: 0.1
            stop_iter: 150000
            below_baseline_penalty_weight: 1.0
            above_baseline_penalty_weight: 0.1
        optim: SGD
        ```

    Note: Default config in config/network/node_config/bias/bias.yaml
        ```yaml
        type: RestingPotential
        groupby:
            - type
        initial_dist: Normal
        mode: sample
        requires_grad: true
        seed: 0
        mean: 0.5
        std: 0.05
        symmetric: []
        penalize:
            activity: true
        ```

    Attributes:
        config (Namespace): Penalty configuration.
        network (Network): The neural network.
        central_cells_index (np.ndarray): Index of central cells.
        parameter_config (Namespace): Configuration for parameter penalties.
        activity_penalty (float): Penalty for activity.
        activity_baseline (float): Baseline for activity.
        activity_penalty_stop_iter (int): Iteration to stop activity penalty.
        below_baseline_penalty_weight (float): Weight for below baseline penalty.
        above_baseline_penalty_weight (float): Weight for above baseline penalty.
        parameter_optim (torch.optim.Optimizer): Optimizer for parameter penalties.
        activity_optim (torch.optim.Optimizer): Optimizer for activity penalties.
        optimizers (Dict[str, torch.optim.Optimizer]): Dictionary of optimizers.
        param_list_func_pen (list): List of parameters for function penalties.
        param_list_act_pen (list): List of parameters for activity penalties.

    Examples:
        Example configurations passed to the network object:

        ```python
        # Example 1: Penalize the resting potential of all cell types.
        bias = Namespace(
            ... (other parameters)
            penalize=Namespace(activity=True),
        )

        # Example 2: add a weight decay penalty to all synapse strengths.
        syn_strength = Namespace(
            ... (other parameters)
            penalize=Namespace(function="weight_decay", kwargs=dict(lambda=1e-3,)),
        )
        ```
    """

    def __init__(self, penalty: Namespace, network: Network):
        self.config = penalty
        self.network = network
        self.central_cells_index = self.network.connectome.central_cells_index[:]

        self.parameter_config = self.get_network_param_penalty_configs()
        self.optim_class = getattr(torch.optim, getattr(self.config, "optim", "SGD"))
        self.parameter_optim, self.activity_optim = None, None
        self.init_optim()
        self.init_hparams()

    def get_network_param_penalty_configs(self) -> Namespace:
        """Returns a dictionary of all network parameters configured to be penalized."""
        node_config = Namespace({
            "nodes_" + k: v.pop("penalize", None)
            for k, v in self.network.config.node_config.deepcopy().items()
        })
        edge_config = Namespace({
            "edges_" + k: v.pop("penalize", None)
            for k, v in self.network.config.edge_config.deepcopy().items()
        })
        return valfilter(
            lambda v: v is not None,
            Namespace(**node_config, **edge_config),
            factory=Namespace,
        )

    def init_optim(self) -> None:
        """Initialize the individual optimizer instances with the correct set of
        parameters.
        """
        self.optimizers = {}
        self.param_list_func_pen = []
        self.param_list_act_pen = []

        # collect the parameters that need to be penalized
        # either by a function or by activity
        for name, config in self.parameter_config.items():
            if "function" in config and any(list(config.kwargs.values())):
                self.param_list_func_pen.append(name)
            if getattr(config, "activity", False):
                self.param_list_act_pen.append(name)

        if self.param_list_func_pen:
            self.parameter_optim = self.optim_class(
                (getattr(self.network, param) for param in self.param_list_func_pen),
                lr=1e-3,
            )  # LR is overwritten by scheduler.
            self.optimizers.update(dict(parameter_optim=self.parameter_optim))

        if self.param_list_act_pen:
            self.activity_optim = self.optim_class(
                (getattr(self.network, param) for param in self.param_list_act_pen),
                lr=1e-3,
            )  # LR is overwritten by scheduler.
            self.optimizers.update(dict(activity_optim=self.activity_optim))

    def init_hparams(self) -> None:
        """Initialize the hyperparameters for the activity penalty."""
        config = self.config.get("activity_penalty", Namespace())

        # collecting activity penalty parameters
        (
            self.activity_penalty,
            self.activity_baseline,
            self.activity_penalty_stop_iter,
            self.below_baseline_penalty_weight,
            self.above_baseline_penalty_weight,
        ) = (
            config.get("activity_penalty", None),
            config.get("activity_baseline", None),
            config.get("stop_iter", None),
            config.get("below_baseline_penalty_weight", None),
            config.get("above_baseline_penalty_weight", None),
        )

        if (
            not any((
                self.activity_penalty,
                self.activity_baseline,
                self.below_baseline_penalty_weight,
                self.above_baseline_penalty_weight,
            ))
            and self.param_list_act_pen
        ):
            raise ValueError(
                "Activity penalty is enabled but no activity penalty parameters are "
                "set."
            )

    def __repr__(self):
        return (
            f"Penalty("
            f"parameter_config={self.parameter_config}, "
            f"activity_penalty={self.activity_penalty}, "
            f"activity_baseline={self.activity_baseline}, "
            f"activity_penalty_stop_iter={self.activity_penalty_stop_iter}, "
            f"below_baseline_penalty_weight={self.below_baseline_penalty_weight}, "
            f"above_baseline_penalty_weight={self.above_baseline_penalty_weight}, "
            f"optim_class={self.optim_class}"
            f")"
        )

    def __call__(self, activity: torch.Tensor, iteration: int) -> None:
        """Run all configured penalties.

        Args:
            activity: Network activity.
            iteration: Current iteration.
        """
        if self.parameter_optim:
            self.param_penalty_step()
        if self.activity_optim:
            if (
                self.activity_penalty_stop_iter is None
                or iteration < self.activity_penalty_stop_iter
            ):
                self.activity_penalty_step(activity, retain_graph=False)
            else:
                self.activity_optim = None

    def _chkpt(self) -> dict:
        """Returns a dictionary of all state dicts of all optimizer instances."""
        _chkpt = {}
        for key, optim in self.optimizers.items():
            if optim is not None:
                _chkpt[key] = optim.state_dict()
        return _chkpt

    def param_penalty_step(self) -> None:
        """Apply all the penalties on the individual parameters."""
        self.parameter_optim.zero_grad()
        penalty = 0
        for param, config in self.parameter_config.items():
            if getattr(config, "function", False):
                penalty += getattr(self, config.function)(param, config)
        penalty.backward()
        self.parameter_optim.step()
        self.network.clamp()

    def activity_penalty_step(
        self, activity: torch.Tensor, retain_graph: bool = True
    ) -> None:
        """Penalizes parameters tracked in activity_optim for too high or low activity.

        Encourages the nodes to have a higher or lower temporal mean activity,
        remedying dead or overactive neurons.

        Note:
            This assumes that the central cells are representative for all cells because
            of shared parameters across the cell types, which makes this reasonably
            efficient.

        Args:
            activity: Network activity of shape (n_samples, n_frames, n_nodes).
            retain_graph: Whether to retain the computation graph.
        """
        self.activity_optim.zero_grad()
        n_samples, n_frames, n_nodes = activity.shape
        # the temporal average activity of the central nodes after a couple of frames
        # to avoid the initial transient response
        activity_mean = activity[:, n_frames // 4 :, self.central_cells_index].mean(
            dim=1
        )  # (n_samples, n_node_types)
        penalty = (
            self.activity_penalty
            * (
                asymmetric_weighting(
                    self.activity_baseline - activity_mean,
                    self.below_baseline_penalty_weight,
                    self.above_baseline_penalty_weight,
                )
                ** 2
            ).mean()
        )
        penalty.backward(retain_graph=retain_graph)
        self.activity_optim.step()
        self.network.clamp()

    def weight_decay(self, param: str, config: Namespace) -> torch.Tensor:
        """Adds weight decay to the loss.

        Warning: Experimental

        Args:
            param: Name of the parameter.
            config: Configuration for the penalty.

        Returns:
            The weight decay penalty.

        """
        w = getattr(self.network, param)
        return config.kwargs["lambda"] * (w**2).sum()

    def prior(self, param: str, config: Namespace) -> torch.Tensor:
        """L2 penalty towards initial values.

        Warning: Experimental

        Args:
            param: Name of the parameter.
            config: Configuration for the penalty.

        Returns:
            The L2 penalty.

        TODO: this might be a convenient but suboptimal implementation when the initial
        values are cast to tensors at each iteration.
        """
        _key = "edge_config" if param.startswith("edges") else "node_config"
        prior = torch.tensor(
            getattr(self.network.config, _key)[
                param.replace("edges_", "").replace("nodes_", "")
            ].value,
            dtype=torch.float32,
        )
        return (
            config.kwargs["lambda"] * ((getattr(self.network, param) - prior) ** 2).sum()
        )


class HyperParamScheduler:
    """Schedules hyperparameters per training iteration.

    Calling the scheduler instance updates the respective hyperparameters per training
    iteration.

    Args:
        scheduler: Scheduler configuration.
        network: The neural network.
        task: The task being solved.
        optimizer: The optimizer.
        penalizer: The penalty object.

    Note: Default config in config/scheduler/scheduler.yaml
        ```yaml
        lr_net:
            function: stepwise
            start: 5.0e-05
            stop: 5.0e-06
            steps: 10
        lr_dec:
            function: stepwise
            start: 5.0e-05
            stop: 5.0e-06
            steps: 10
        lr_pen:
            function: stepwise
            start: ${scheduler.lr_net.start}
            stop: ${scheduler.lr_net.stop}
            steps: 10
        dt:
            function: stepwise
            start: 0.02
            stop: 0.02
            steps: 10
        chkpt_every_epoch: 300
        ```

    Attributes:
        config (Namespace): Scheduler configuration.
        scheduled_params (Namespace): Scheduled parameters.
        network (Network): The neural network.
        task (Task): The task being solved.
        optimizer (torch.optim.Optimizer): The optimizer.
        penalizer (Penalty): The penalty object.
        stop_iter (int): Iteration to stop scheduling.
        _current_iteration (int): Current iteration.
    """

    def __init__(
        self,
        scheduler: Namespace,
        network: Optional[Network],
        task: Optional[Task],
        optimizer: Optional[torch.optim.Optimizer],
        penalizer: Optional[Penalty],
    ):
        self.config = scheduler.deepcopy()
        self.scheduled_params = self.config.deepcopy()
        self.network = network
        self.task = task
        self.optimizer = optimizer
        self.penalizer = penalizer

        self.stop_iter = scheduler.get("sched_stop_iter", self.task.n_iters)
        self._current_iteration = 0

        self.scheduled_params = Namespace()
        for key, param in self.config.items():
            try:
                schedfn_config = SchedulerFunction(**param)
                logging.info("Init schedule for %s", key)
            except TypeError:
                # lazy way to skip the parameter if it's not a SchedulerFunction
                continue

            # these are the parameters that are scheduled
            param.array = getattr(self, schedfn_config.function)(
                self.stop_iter,
                self.task.n_iters,
                param.start,
                param.stop,
                param.steps,
            )
            self.scheduled_params[key] = param

    def __call__(self, iteration: int) -> None:
        """Update hyperparameters for the given iteration.

        Args:
            iteration: Current iteration.
        """
        self._current_iteration = iteration
        for key, param in self.scheduled_params.items():
            try:
                setattr(self, key, param.array[iteration])
            except IndexError as e:
                if iteration >= self.stop_iter:
                    setattr(self, key, param.array[-1])
                else:
                    raise e
        logging.info(self)

    def __repr__(self):
        return "Scheduler. Iteration: {}/{}.\nCurrent values: {}.".format(
            self._current_iteration,
            self.task.n_iters,
            self._params(),
        )

    def _params(self) -> dict:
        """Get current parameter values.

        Returns:
            A dictionary of current parameter values.
        """
        params = {}
        for key, _param in self.scheduled_params.items():
            value = getattr(self, key)
            params[key] = value
        return params

    # -------- Setter methods called automatically by the scheduler

    @property
    def dt(self) -> float:
        return self.task.dataset.dt

    @dt.setter
    def dt(self, value: float) -> None:
        self.task.dataset.dt = value

    @property
    def lr_net(self) -> float:
        if self.optimizer is None:
            return
        return self.optimizer.param_groups[0]["lr"]

    @lr_net.setter
    def lr_net(self, value: float) -> None:
        if self.optimizer is None:
            return
        self.optimizer.param_groups[0]["lr"] = value

    @property
    def lr_dec(self) -> list:
        if self.optimizer is None:
            return
        return [param_group["lr"] for param_group in self.optimizer.param_groups[1:]]

    @lr_dec.setter
    def lr_dec(self, value: float) -> None:
        if self.optimizer is None:
            return
        for param_group in self.optimizer.param_groups[1:]:
            param_group["lr"] = value

    @property
    def lr_pen(self) -> list:
        if self.penalizer is None:
            return
        return [
            param_group["lr"]
            for optim in self.penalizer.optimizers.values()
            for param_group in optim.param_groups
        ]

    @lr_pen.setter
    def lr_pen(self, value: float) -> None:
        if self.penalizer is None:
            return
        for optim in self.penalizer.optimizers.values():
            if optim is not None:
                for param_group in optim.param_groups:
                    param_group["lr"] = value

    @property
    def relu_leak(self) -> float:
        if self.network is None:
            return
        return getattr(self.network.dynamics.activation, "negative_slope", None)

    @relu_leak.setter
    def relu_leak(self, value: float) -> None:
        if self.network is None:
            return
        if hasattr(self.network.dynamics.activation, "negative_slope"):
            self.network.dynamics.activation.negative_slope = value

    @property
    def activity_penalty(self) -> float:
        if self.penalizer is None:
            return
        return self.penalizer.activity_penalty

    @activity_penalty.setter
    def activity_penalty(self, value: float) -> None:
        if self.penalizer is None:
            return
        self.penalizer.activity_penalty = value

    # -------- Decay Options

    @staticmethod
    def linear(
        stop_iter: int, n_iterations: int, start: float, stop: float, steps: int
    ) -> np.ndarray:
        """Generate a linear schedule from start to stop value."""
        f = np.linspace(start, stop, stop_iter)
        return np.pad(f, (0, n_iterations - len(f) + 1), constant_values=stop)

    @staticmethod
    def stepwise(
        stop_iter: int, n_iterations: int, start: float, stop: float, steps: int
    ) -> np.ndarray:
        """Generate a stepwise schedule from start to stop value."""
        f = np.linspace(start, stop, steps).repeat(stop_iter / steps)
        return np.pad(f, (0, n_iterations - len(f) + 1), constant_values=stop)

    @staticmethod
    def stepwise_2ndhalf(
        stop_iter: int, n_iterations: int, start: float, stop: float, steps: int
    ) -> np.ndarray:
        """Generate a stepwise schedule that decays in the second half of iterations."""
        f = np.linspace(start, stop, steps).repeat((stop_iter / 2) / steps)
        return np.pad(f, (n_iterations - len(f) + 1, 0), constant_values=start)

    @staticmethod
    def stepwise_half(
        stop_iter: int, n_iterations: int, start: float, stop: float, steps: int
    ) -> np.ndarray:
        """Generate a stepwise schedule that decays in the first half of iterations."""
        f = np.linspace(start, stop, steps).repeat((stop_iter / 2) / steps)
        return np.pad(f, (0, n_iterations - len(f) + 1), constant_values=stop)

    @staticmethod
    def steponential(
        stop_iter: int, n_iterations: int, start: float, stop: float, steps: int
    ) -> np.ndarray:
        """Generate an exponential stepwise schedule from start to stop value."""
        x = (1 / stop) ** (1 / steps)
        values = start / x ** np.arange(steps)
        f = values.repeat(stop_iter / steps)
        return np.pad(f, (0, n_iterations - len(f) + 1), constant_values=values[-1])

    @staticmethod
    def steponential_inv(
        stop_iter: int, n_iterations: int, start: float, stop: float, steps: int
    ) -> np.ndarray:
        """Generate an inverse exponential stepwise schedule."""
        _start = steps
        _stop = 0
        x = 1 / _stop
        values = _start / x ** np.arange(steps)
        f = values.repeat(stop_iter / steps)
        return np.pad(f, (0, n_iterations - len(f) + 1), constant_values=values[-1])

    @staticmethod
    def exponential(
        stop_iter: int, n_iterations: int, start: float, stop: float, steps: int
    ) -> np.ndarray:
        """Generate an exponential schedule from start to stop value."""
        tau = -stop_iter / (np.log(stop + 1e-15) - np.log(start))
        f = start * np.exp(-np.arange(stop_iter) / tau)
        return np.pad(f, (0, n_iterations - len(f) + 1), constant_values=stop)

    @staticmethod
    def exponential_half(
        stop_iter: int, n_iterations: int, start: float, stop: float, steps: int
    ) -> np.ndarray:
        """Generate exponential schedule that decays in the first half of iterations."""
        tau = -int((stop_iter / 2)) / (np.log(stop) - np.log(start))
        f = start * np.exp(-np.arange(int(stop_iter / 2)) / tau)
        return np.pad(f, (0, n_iterations - len(f) + 1), constant_values=stop)


@dataclass
class SchedulerFunction:
    """Configuration for a scheduler function.

    Attributes:
        start (float): Start value.
        stop (float): Stop value.
        steps (int): Number of steps.
        function (str): Name of the scheduling function.
    """

    start: float
    stop: float
    steps: int
    function: str



"""CLI entry point and pipeline manager for ensemble operations.


Example usage:
    ```bash
    # Run a pipeline of operations
    flyvis \
        --ensemble_id 0001 \
        --task_name flow \
        train \
        validate \
        record \
        analysis \
        notebook_per_model \
        notebook_per_ensemble
    ```
"""

# TODO: could look into click and typer for better CLI handling but this works for now.

import argparse
import subprocess
import sys
from importlib import resources
from pathlib import Path
from typing import List, Tuple

SCRIPTS_DIR = resources.files('flyvis_cli')

SCRIPT_COMMANDS = {
    "train": SCRIPTS_DIR / "training/train.py",
    "train-single": SCRIPTS_DIR / "training/train_single.py",
    "validate": SCRIPTS_DIR / "validation/validate.py",
    "val-single": SCRIPTS_DIR / "validation/val_single.py",
    "record": SCRIPTS_DIR / "analysis/record.py",
    "synthetic-recordings-single": SCRIPTS_DIR
    / "analysis/synthetic_recordings_single.py",
    "analysis": SCRIPTS_DIR / "analysis/analysis.py",
    "ensemble-analysis": SCRIPTS_DIR / "analysis/ensemble_analysis.py",
    "notebook-per-model": SCRIPTS_DIR / "analysis/notebook_per_model.py",
    "notebook-per-ensemble": SCRIPTS_DIR / "analysis/notebook_per_ensemble.py",
    "notebook": SCRIPTS_DIR / "analysis/notebook.py",
    "download-pretrained": SCRIPTS_DIR / "download_pretrained_models.py",
    "init-config": SCRIPTS_DIR / "init_config.py",
}


def run_script(script_path: Path, args: List[str]) -> None:
    """
    Run a Python script with the given arguments.

    Args:
        script_path: Path to the script to run.
        args: List of command-line arguments to pass to the script.
    """
    cmd = [sys.executable, str(script_path)] + args
    try:
        subprocess.run(cmd, check=True)
    except subprocess.CalledProcessError as e:
        print(f"Error running {script_path}: {e}", file=sys.stderr)
        sys.exit(1)


def filter_args(
    argv: List[str], allowed_commands: List[str]
) -> Tuple[List[str], List[str]]:
    """
    Filter out commands from command line arguments.

    Args:
        argv: List of command line arguments (typically sys.argv[1:])
        commands: List of commands to filter out

    Returns:
        List of arguments with commands removed
    """
    selected_commands = []
    other_args = []
    for arg in argv:
        if arg in allowed_commands:
            selected_commands.append(arg)
        else:
            other_args.append(arg)

    return selected_commands, other_args


def handle_help_request(argv: List[str]) -> bool:
    """
    Check if help is requested for a specific command and handle it.

    Args:
        argv: Command line arguments (typically sys.argv)

    Returns:
        True if help was handled, False otherwise
    """
    if len(argv) > 1 and argv[1] in SCRIPT_COMMANDS:
        help_flags = {"-h", "--help"}
        if any(flag in argv[2:] for flag in help_flags):
            command = argv[1]
            run_script(SCRIPT_COMMANDS[command], argv[2:])
            return True
    return False


def main():
    # Handle help requests first
    if handle_help_request(sys.argv):
        return 0

    parser = argparse.ArgumentParser(
        description=(
            "Run flyvis pipelines or individual scripts with compute cloud options."
        ),
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "--ensemble_id", type=int, required=False, help="Id of the ensemble, e.g. 0045."
    )
    parser.add_argument(
        "--task_name", type=str, required=False, help="Name given to the task, e.g. flow."
    )
    parser.add_argument(
        "commands",
        nargs="+",
        # choices=list(SCRIPT_COMMANDS.keys()),
        help="Commands to run in order.",
        metavar="command",
    )

    parser.epilog = """
Runs multiple operations on an ensemble of models.
This is to pipeline jobs on the compute cloud.
Each command corresponds to a script that launches required jobs.

{}

All additional arguments are passed directly to the respective scripts.
For detailed help on each command, run: flyvis <command> --help
""".format('\n'.join(f"{cmd:<20} : Runs {path}" for cmd, path in SCRIPT_COMMANDS.items()))

    # args, remaining = parser.parse_known_args()

    # Filter out commands from arguments
    selected_commands, other_args = filter_args(
        sys.argv[1:], list(SCRIPT_COMMANDS.keys())
    )
    if not selected_commands:
        parser.print_help()
        return 1

    for command in selected_commands:
        run_script(SCRIPT_COMMANDS[command], other_args)
    return 0


if __name__ == "__main__":
    sys.exit(main())



"""Ensemble of trained networks."""

from __future__ import annotations

import logging
from functools import wraps
from pathlib import Path
from typing import Callable, Iterable, List, Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import colormaps as cm
from torch import nn

import flyvis
from flyvis.analysis.flash_responses import flash_response_index, plot_fris
from flyvis.analysis.moving_bar_responses import (
    direction_selectivity_index,
    dsi_violins_on_and_off,
)
from flyvis.analysis.visualization import plots
from flyvis.connectome import get_avgfilt_connectome
from flyvis.utils.chkpt_utils import (
    best_checkpoint_default_fn,
    recover_network,
    resolve_checkpoints,
)

from .directories import EnsembleDir
from .ensemble import Ensemble
from .network import Network

logging = logging.getLogger(__name__)

__all__ = ["EnsembleView"]


class EnsembleView(Ensemble):
    """A view of an ensemble of trained networks.

    This class extends the Ensemble class with visualization and analysis methods.

    Args:
        path: Path to the ensemble directory or an existing Ensemble object.
        network_class: The network class to use for instantiation.
        root_dir: Root directory for results.
        connectome_getter: Function to get the connectome.
        checkpoint_mapper: Function to resolve checkpoints.
        best_checkpoint_fn: Function to select the best checkpoint.
        best_checkpoint_fn_kwargs: Keyword arguments for best_checkpoint_fn.
        recover_fn: Function to recover the network.
        try_sort: Whether to try sorting the ensemble.

    Attributes:
        Inherits all attributes from the Ensemble class.
    """

    def __init__(
        self,
        path: Union[str, Path, Iterable, EnsembleDir, Ensemble],
        network_class: nn.Module = Network,
        root_dir: Path = flyvis.results_dir,
        connectome_getter: Callable = get_avgfilt_connectome,
        checkpoint_mapper: Callable = resolve_checkpoints,
        best_checkpoint_fn: Callable = best_checkpoint_default_fn,
        best_checkpoint_fn_kwargs: dict = {
            "validation_subdir": "validation",
            "loss_file_name": "epe",
        },
        recover_fn: Callable = recover_network,
        try_sort: bool = False,
    ):
        init_args = (
            path,
            network_class,
            root_dir,
            connectome_getter,
            checkpoint_mapper,
            best_checkpoint_fn,
            best_checkpoint_fn_kwargs,
            recover_fn,
            try_sort,
        )
        if isinstance(path, Ensemble):
            init_args = path._init_args
        super().__init__(*init_args)

    @wraps(plots.loss_curves)
    def training_loss(self, **kwargs) -> Tuple[plt.Figure, plt.Axes]:
        """Plot training loss curves for the ensemble.

        Args:
            **kwargs: Additional keyword arguments to pass to plots.loss_curves.

        Returns:
            A tuple containing the matplotlib Figure and Axes objects.
        """
        task_error = self.task_error()
        losses = np.array([nv.dir.loss[:] for nv in self.values()])
        return plots.loss_curves(
            losses,
            cbar=True,
            colors=task_error.colors,
            cmap=task_error.cmap,
            norm=task_error.norm,
            xlabel="iterations",
            ylabel="training loss",
            **kwargs,
        )

    @wraps(plots.loss_curves)
    def validation_loss(
        self,
        validation_subdir: Optional[str] = None,
        loss_file_name: Optional[str] = None,
        **kwargs,
    ) -> Tuple[plt.Figure, plt.Axes]:
        """Plot validation loss curves for the ensemble.

        Args:
            validation_subdir: Subdirectory containing validation data.
            loss_file_name: Name of the loss file.
            **kwargs: Additional keyword arguments to pass to plots.loss_curves.

        Returns:
            A tuple containing the matplotlib Figure and Axes objects.
        """
        task_error = self.task_error()
        losses = self.validation_losses(validation_subdir, loss_file_name)
        return plots.loss_curves(
            losses,
            cbar=True,
            colors=task_error.colors,
            cmap=task_error.cmap,
            norm=task_error.norm,
            xlabel="checkpoints",
            ylabel="validation loss",
            **kwargs,
        )

    @wraps(plots.histogram)
    def task_error_histogram(self, **kwargs) -> Tuple[plt.Figure, plt.Axes]:
        """Plot a histogram of the validation losses of the ensemble.

        Args:
            **kwargs: Additional keyword arguments to pass to plots.histogram.

        Returns:
            A tuple containing the matplotlib Figure and Axes objects.
        """
        losses = self.min_validation_losses()
        return plots.histogram(
            losses, xlabel="task error", ylabel="number models", **kwargs
        )

    @wraps(plots.violins)
    def node_parameters(
        self, key: str, max_per_ax: int = 34, **kwargs
    ) -> Tuple[plt.Figure, List[plt.Axes]]:
        """Plot violin plots of node parameters for the ensemble.

        Args:
            key: The parameter key to plot.
            max_per_ax: Maximum number of violins per axis.
            **kwargs: Additional keyword arguments to pass to plots.violins.

        Returns:
            A tuple containing the matplotlib Figure and a list of Axes objects.
        """
        parameters = self.parameters()[f"nodes_{key}"]
        parameter_keys = self.parameter_keys()[f"nodes_{key}"]
        return plots.violins(
            parameter_keys, parameters, ylabel=key, max_per_ax=max_per_ax, **kwargs
        )

    @wraps(plots.violins)
    def edge_parameters(
        self, key: str, max_per_ax: int = 120, **kwargs
    ) -> Tuple[plt.Figure, List[plt.Axes]]:
        """Plot violin plots of edge parameters for the ensemble.

        Args:
            key: The parameter key to plot.
            max_per_ax: Maximum number of violins per axis.
            **kwargs: Additional keyword arguments to pass to plots.violins.

        Returns:
            A tuple containing the matplotlib Figure and a list of Axes objects.
        """
        parameters = self.parameters()[f"edges_{key}"]
        parameter_keys = self.parameter_keys()[f"edges_{key}"]
        variable_names = np.array([
            f"{source}->{target}" for source, target in parameter_keys
        ])
        return plots.violins(
            variable_names,
            variable_values=parameters,
            ylabel=key,
            max_per_ax=max_per_ax,
            **kwargs,
        )

    @wraps(plots.heatmap)
    def dead_or_alive(self, **kwargs) -> Tuple[plt.Figure, plt.Axes]:
        """Plot a heatmap of dead cells in the ensemble.

        Args:
            **kwargs: Additional keyword arguments to pass to plots.heatmap.

        Returns:
            A tuple containing the matplotlib Figure and Axes objects.
        """
        responses = self.naturalistic_stimuli_responses()
        dead_count = (responses['responses'].values < 0).all(axis=(1, 2))
        return plots.heatmap(
            dead_count,
            ylabels=np.arange(len(self)),
            xlabels=responses.cell_type.values,
            size_scale=15,
            cbar=False,
            **kwargs,
        )

    @wraps(plot_fris)
    def flash_response_index(
        self, cell_types: Optional[List[str]] = None, **kwargs
    ) -> Tuple[plt.Figure, plt.Axes]:
        """Plot the flash response indices of the ensemble.

        Args:
            cell_types: List of cell types to include. If None, all cell types are used.
            **kwargs: Additional keyword arguments to pass to plot_fris.

        Returns:
            A tuple containing the matplotlib Figure and Axes objects.
        """
        responses = self.flash_responses()
        fris = flash_response_index(responses, radius=6)
        if cell_types is not None:
            requested_cell_types = cell_types
            fris = fris.custom.where(cell_type=cell_types)
            cell_types = fris.cell_type.values
            kwargs.setdefault("sorted_type_list", requested_cell_types)
        else:
            cell_types = fris.cell_type.values
        task_error = self.task_error()
        best_index = np.argmin(task_error.values)
        return plot_fris(
            fris.values,
            cell_types,
            scatter_best=True,
            scatter_best_index=best_index,
            scatter_best_color=cm.get_cmap("Blues")(1.0),
            **kwargs,
        )

    @wraps(dsi_violins_on_and_off)
    def direction_selectivity_index(
        self, **kwargs
    ) -> Tuple[plt.Figure, Tuple[plt.Axes, plt.Axes]]:
        """Plot the direction selectivity indices of the ensemble.

        Args:
            **kwargs: Additional keyword arguments to pass to plot_dsis.

        Returns:
            A tuple containing the matplotlib Figure and a tuple of Axes objects.
        """
        responses = self.moving_edge_responses()
        dsis = direction_selectivity_index(responses)
        task_error = self.task_error()
        best_index = np.argmin(task_error.values)
        return dsi_violins_on_and_off(
            dsis,
            responses.cell_type,
            bold_output_type_labels=False,
            figsize=[10, 1.2],
            color_known_types=True,
            fontsize=6,
            scatter_best_index=best_index,
            scatter_best_color=cm.get_cmap("Blues")(1.0),
            **kwargs,
        )



import argparse
import inspect
import os
import sys
import warnings
from importlib import resources
from typing import Any, Dict, List, Optional, Union

import hydra
from colorama import Fore, Style, init
from datamate import Namespace, namespacify
from hydra.core.global_hydra import GlobalHydra
from omegaconf import DictKeyType, OmegaConf, errors

CONFIG_PATH = str(resources.files("flyvis") / "config")

init(autoreset=True)  # Initialize colorama


def get_default_config(
    overrides: List[str],
    path: str = "../../config/solver.yaml",
    as_namespace: bool = True,
) -> Union[Dict[DictKeyType, Any], List[Any], None, str, Any, Namespace]:
    """
    Get the default configuration using Hydra.

    Args:
        overrides: List of configuration overrides.
        path: Path to the configuration file.
        as_namespace: Whether to return a namespaced configuration or the
            OmegaConf object.

    Returns:
        The configuration object.

    Note:
        Expected overrides are:
        - task_name
        - network_id
    """

    config = get_config_from_file(path, overrides, resolve=True)
    if as_namespace:
        return namespacify(config)
    return config


def get_config_from_file(
    path: str,
    overrides: List[str] = [],
    resolve: bool = False,
    throw_on_missing: bool = False,
) -> Union[Dict[DictKeyType, Any], List[Any], None, str, Any]:
    """
    Get the configuration from a file.

    Args:
        path: Path to the configuration file (absolute or relative).
        overrides: List of configuration overrides.
        resolve: Whether to resolve the configuration.
        throw_on_missing: Whether to throw an error if a key is missing.

    Returns:
        The configuration object.
    """
    # Get absolute paths
    caller_frame = inspect.stack()[1]
    caller_file = caller_frame.filename
    caller_dir = os.path.dirname(os.path.abspath(caller_file))

    abs_config_path = os.path.join(caller_dir, path) if not os.path.isabs(path) else path

    config_name = os.path.basename(abs_config_path).replace(".yaml", "")
    config_dir = os.path.dirname(abs_config_path)

    # Calculate relative path from caller to config directory
    rel_config_dir = os.path.relpath(config_dir, caller_dir)
    GlobalHydra.instance().clear()
    hydra.initialize(config_path=rel_config_dir, version_base=None)
    config = hydra.compose(config_name=config_name, overrides=overrides)
    config = OmegaConf.to_container(
        config, resolve=resolve, throw_on_missing=throw_on_missing
    )
    GlobalHydra.instance().clear()
    return config


class HybridArgumentParser(argparse.ArgumentParser):
    """
    Hybrid argument parser that can parse unknown arguments in basic key=value style.

    Attributes:
        hybrid_args: Dictionary of hybrid arguments with their requirements and
            help texts.
        allow_unrecognized: Whether to allow unrecognized arguments.
        drop_disjoint_from: Path to a configuration file that can be used to filter
            out arguments that are present in the command line arguments but not in
            the configuration file. This is to pass through arguments through multiple
            scripts as hydra does not support this.

    Args:
        hybrid_args: Dictionary of hybrid arguments with their requirements and
            help texts.
        allow_unrecognized: Whether to allow unrecognized arguments.
    """

    def __init__(
        self,
        *args: Any,
        hybrid_args: Optional[Dict[str, Dict[str, Any]]] = None,
        allow_unrecognized: bool = True,
        drop_disjoint_from: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.hybrid_args = hybrid_args or {}
        self.allow_unrecognized = allow_unrecognized
        self.drop_disjoint_from = drop_disjoint_from
        self._add_hybrid_args_to_help()

    def _add_hybrid_args_to_help(self) -> None:
        """Add hybrid arguments to the help message."""
        if self.hybrid_args:
            hybrid_group = self.add_argument_group('Hybrid Arguments')
            for arg, config in self.hybrid_args.items():
                help_text = config.get('help', '')
                required = config.get('required', False)
                arg_type = config.get('type', None)
                arg_help = f"{arg}=value: {help_text}"
                if arg_type:
                    arg_help += f" (type: {arg_type.__name__})"
                if required:
                    arg_help += " (Required)"
                hybrid_group.add_argument(f"--{arg}", help=arg_help, required=False)

    def parse_with_hybrid_args(
        self,
        args: Optional[List[str]] = None,
        namespace: Optional[argparse.Namespace] = None,
    ) -> argparse.Namespace:
        """
        Parse arguments and set hybrid arguments as attributes in the namespace.

        Args:
            args: List of arguments to parse.
            namespace: Namespace to populate with parsed arguments.

        Returns:
            Namespace with parsed arguments.

        Raises:
            argparse.ArgumentError: If required arguments are missing or invalid
                values are provided.
        """
        if args is None:
            args = sys.argv[1:]

        args_for_parser = []
        key_value_args = []

        # Separate key=value pairs from other arguments
        for arg in args:
            if '=' in arg and not arg.startswith('-'):
                key_value_args.append(arg)
            else:
                args_for_parser.append(arg)

        # Parse the known arguments
        args, unknown_args = self.parse_known_args(args_for_parser, namespace)

        # Combine key_value_args with unknown_args for processing
        all_unknown_args = key_value_args + unknown_args

        argv = []
        for arg in all_unknown_args:
            if ":" in arg and "=" in arg:
                keytype, value = arg.split("=")
                key, astype = keytype.split(":")
                try:
                    if value.lower() in ["true", "1", 'yes'] and astype == "bool":
                        setattr(args, key, True)
                    elif value.lower() in ["false", "0", 'no'] and astype == "bool":
                        setattr(args, key, False)
                    else:
                        setattr(args, key, safe_cast(value, astype))
                except (ValueError, TypeError):
                    self.error(
                        f"Invalid type '{astype}' or value '{value}' for argument {key}"
                    )
            elif "=" in arg:
                key, value = arg.split("=", 1)
                if key in self.hybrid_args and 'type' in self.hybrid_args[key]:
                    arg_type = self.hybrid_args[key]['type']
                    try:
                        typed_value = arg_type(value)
                        setattr(args, key, typed_value)
                    except ValueError:
                        self.error(
                            f"Invalid {arg_type.__name__} value '{value}' "
                            f"for argument {key}"
                        )
                else:
                    setattr(args, key, value)
            else:
                argv.append(arg)

        # Apply type conversion for arguments parsed by argparse
        for arg, config in self.hybrid_args.items():
            if (
                hasattr(args, arg)
                and config.get('type')
                and getattr(args, arg) is not None
            ):
                setattr(args, arg, config['type'](getattr(args, arg)))

        # Check for required arguments
        missing_required = []
        for arg, config in self.hybrid_args.items():
            if config.get('required', False) and getattr(args, arg) is None:
                missing_required.append(arg)

        if missing_required:
            self.error(
                f"The following required arguments are missing: "
                f"{', '.join(missing_required)}"
            )

        if argv and not self.allow_unrecognized:
            msg = "unrecognized arguments: %s"
            self.error(msg % " ".join(argv))

        if self.drop_disjoint_from:
            args = self._filter_args_based_on_config(args)

        return args

    def hydra_argv(self) -> List[str]:
        hybrid_args = self.parse_with_hybrid_args()
        return [
            f"{key}={value}" for key, value in vars(hybrid_args).items() if ":" not in key
        ]

    def get_registered_args(self) -> List[str]:
        """
        Get a list of all argument names that were registered using add_argument.

        Returns:
            List of argument names (without the -- prefix)
        """
        return [
            action.dest
            for action in self._actions
            if action.dest != "help"  # Exclude the default help action
        ]

    def _filter_args_based_on_config(
        self, args: argparse.Namespace
    ) -> argparse.Namespace:
        """
        Filter arguments based on the Hydra config file specified in drop_disjoint_from.

        Args:
            args: Namespace containing all parsed arguments.

        Returns:
            Filtered Namespace with only arguments present in the config or with
                Hydra syntax.
        """
        if not self.drop_disjoint_from:
            return args

        config = OmegaConf.create(
            get_config_from_file(self.drop_disjoint_from, resolve=False)
        )

        filtered_args = argparse.Namespace()
        registered_args = self.get_registered_args()

        for arg, value in vars(args).items():
            if (
                self._is_in_config(arg, config)
                or arg.startswith('+')
                or arg.startswith('++')
                or arg.startswith('~')
            ):
                setattr(filtered_args, arg, value)
            elif arg not in registered_args:
                warnings.warn(
                    f"{Fore.YELLOW}Argument {Style.BRIGHT}{arg}={value}"
                    f"{Style.RESET_ALL}{Fore.YELLOW} "
                    f"does not affect the hydra config because it is not present in "
                    f"the config file {Style.BRIGHT}{self.drop_disjoint_from}"
                    f"{Style.RESET_ALL}{Fore.YELLOW}. "
                    f"This may be unintended, like a typo, or intended, like a "
                    f"hydra-style argument passed through to another script. "
                    f"Check script docs and config file "
                    f"for clarification.{Style.RESET_ALL}",
                    stacklevel=2,
                )

        return filtered_args

    def _is_in_config(self, arg: str, config: Any) -> bool:
        """
        Check if an argument exists in the config, including nested structures.

        Args:
            arg: The argument to check.
            config: The configuration object or sub-object.

        Returns:
            True if the argument is found in the config, False otherwise.
        """
        try:
            return OmegaConf.select(config, arg, throw_on_missing=True) is not None
        except errors.MissingMandatoryValue:
            return True


def parse_kwargs_to_dict(values: List[str]) -> argparse.Namespace:
    """
    Parse a list of key-value pairs into a dictionary.

    Args:
        values: List of key-value pairs in the format "key=value".

    Returns:
        Namespace object with parsed key-value pairs as attributes.
    """
    kwargs = argparse.Namespace()
    for value in values:
        key, value = value.split("=")
        setattr(kwargs, key, value)
    return kwargs


def safe_cast(value: str, type_name: str) -> Union[int, float, bool, str]:
    """
    Safely cast a string value to a specified type.

    Args:
        value: The string value to cast.
        type_name: The name of the type to cast to.

    Returns:
        The casted value.

    Note:
        Supports casting to int, float, bool, and str.
    """
    if type_name == 'int':
        return int(value)
    elif type_name == 'float':
        return float(value)
    elif type_name == 'bool':
        return value.lower() in ('true', 'yes', '1', 'on')
    else:
        return value  # Default to string


if __name__ == "__main__":
    parser = HybridArgumentParser(
        hybrid_args={"abc": {"type": str, "required": True}},
        drop_disjoint_from="../../config/solver.yaml",
        allow_unrecognized=True,
    )
    args, kwargs = parser.parse_known_args()
    print(args)
    print(kwargs)
    args = parser.parse_with_hybrid_args()
    print(args)
    kwargs = ["=".join(arg) for arg in vars(args).items()]
    print(kwargs)
